import inspect
import numpy as np
import pandas as pd
from scipy.stats import chi2
from pandas import DataFrame, Series
from typing import Any, List, Union, Optional
from kivi.woe.base import WOEMixin


__all__ = [
    "Chi2Bins"
]


def chi2_square(freq):
    """计算卡方值以检测目标类的频率在给定区间内是否存在显着差异。
    参考: "ChiMerge: DIscretization of Numerical Attributes"

    参数:
    :param freq: 分组与分箱的列联表（频数）。
    :returns: chi2 卡方统计量

    示例：
    >>> freq = np.array([[100, 9], [101, 8]])
    >>> # 计算卡方统计量
    >>> chi2_square(freq)
    """
    R = freq.sum(axis=1).reshape(1, -1)  # 每一个分组的样本数
    C = freq.sum(axis=0).reshape(1, -1).T  # 每一个类别的样本数
    # 期望频率 (加 0.5 避免小数)
    E = np.dot(C, R) / C.sum() + 0.5
    chi2_value = (np.square(freq.T - E) / E).sum()  # 卡方值
    return chi2_value


def BinsChiTest(m, pt_value, pt_column, pt_index):
    """
    对每对 m 个相邻区间执行卡方检验，使用第 i 个区间的索引作为区间对的索引
    """

    adjacent_list = (pt_value[i:i + m, :] for i in range(len(pt_value) - m + 1))
    adjacent_index = np.array([pt_index[i:i + m] for i in range(len(pt_value) - m + 1)])
    chi2_array = np.array([chi2_square(adj) for adj in adjacent_list])
    return adjacent_index, chi2_array


def pivot_table(index, column):
    """ """
    df_pivot = pd.DataFrame(np.vstack([index, column]).T, columns=['index', 'column'])
    df_pivot['cnt'] = 1

    if 'sort' in inspect.getfullargspec(pd.DataFrame.pivot_table).args:
        df_pivot = df_pivot.pivot_table(columns='column', index='index', aggfunc='count', sort=True)
    else:
        df_pivot = df_pivot.pivot_table(columns='column', index='index', aggfunc='count')

    df_pivot.fillna(0, inplace=True)

    pivot_table = df_pivot.values
    column_unique = df_pivot.columns.to_numpy()
    index_unique = df_pivot.index.to_numpy()
    return pivot_table, column_unique, index_unique


def assign_interval_unique(x, boundaries):
    """为 x 中的每个值分配一个距边界的间隔

    Parameters
    ----------
    x: 需要离散化的数据列。

    boundaries: 离散化目标 x 的区间边界值。

    Returns
    -------
    intervals: numpy.ndarray, shape (number of examples,2)
        向右闭合的区间数组。数组的左列和右列分别是左右边界。

    unique_intervals: numpy.ndarray, shape (number of unique intervals,2)
        向右闭合的唯一区间。数组的左列和右列分别是左右边界。
    """
    # Add -inf and inf to the start and end of boundaries
    max_value = max(x)
    boundaries = np.unique(np.concatenate((np.array([-float('inf')]), boundaries, np.array([float('inf')])), axis=0))

    # The max boundary that is smaller than x_i is its lower boundary.
    # The min boundary that is >= than x_i is its upper boundary.
    # Adding equal is because all intervals here are closed to the right.
    boundaries_diff_boolean = x.reshape(1, -1).T > boundaries.reshape(1, -1)
    lowers = np.array([boundaries[b].max() for b in boundaries_diff_boolean])
    uppers = np.array([boundaries[b].min() for b in ~boundaries_diff_boolean])

    # Replace the upper value with inf if it is not smaller then the maximum feature value
    n = x.shape[0]
    uppers = np.where(uppers >= max_value, [float('inf')] * n, uppers)
    # Array of intervals that are closed to the right
    intervals = np.stack((lowers, uppers), axis=1)

    unique_intervals = np.unique(intervals, axis=0)
    return intervals, unique_intervals


def chi_merge_vector(x, y, m=2, confidence_level=0.9, max_intervals=None,
                     min_intervals=2, initial_intervals=100) -> List[Union[int, float]]:
    """合并相似的相邻 m 个区间，直到所有相邻区间彼此显着不同。

    Parameters
    ----------

    x: numpy.array, shape (number of examples,)
        The array of data that need to be discretized.

    y: numpy.array, shape (number of examples,)
        The target array (or dependent variable).

    m: integer, optional(default=2)
        The number of adjacent intervals to compare during chi-squared test.

    confidence_level: float, optional(default=0.9)
        The confidence level to determine the threshold for intervals to
        be considered as different during the chi-square test.

    max_intervals: int, optional(default=None)
        Specify the maximum number of intervals the discretized array will have.
        Sometimes (like when training a scorecard model) fewer intervals are
        prefered. If do not need this option just set it to None.

    min_intervals: int, optional(default=2)
        Specify the mininum number of intervals the discretized array will have.
        If do not need this option just set it to 2.

    initial_intervals: int, optional(default=100)
        The original Chimerge algorithm starts by putting each unique value
        in an interval and merging through a loop. This can be time-consumming
        when sample size is large.
        Set the initial_intervals option to values other than None (like 10 or 100)
        will make the algorithm start at the number of intervals specified (the
        initial intervals are generated using quantiles). This can greatly shorten
        the run time. If do not need this option just set it to None.

    Return
    ------
    """
    # --------------------
    # 初始化阶段
    # --------------------
    n_j = np.unique(y).shape[0]  # 分类类别的数量
    n_i = np.unique(x).shape[0]  # 变量 x 的不同的值
    if (initial_intervals is not None and initial_intervals < n_i and min_intervals < n_i):
        # 离散分段数量小于 n_i 的值，最小分段数量应小于 n_i
        boundaries = np.unique(np.quantile(x, np.arange(0, 1, 1 / initial_intervals)[1:]))
    else:
        boundaries = np.unique(x)

    intervals, unique_intervals = assign_interval_unique(x, boundaries)

    # 当 unique x <= min_intervals 时，返回 X unique 值作为结果
    if n_i <= min_intervals:
        if n_i == 1:
            boundaries = np.array([float('inf')])
        return boundaries  # output the unique upper boundaries of discretized array

    # --------------------
    # 迭代步骤
    # --------------------
    if max_intervals is None:
        max_intervals = n_i

    threshold = chi2.ppf(confidence_level, n_j - 1)  # chi2 阈值
    # 各个分组的样本计数
    pt_value, pt_column, pt_index = pivot_table(intervals[:, 1], y)

    # 对每对 m 个相邻区间执行卡方检验，使用第 i 个区间的索引作为区间对的索引
    adjacent_index, chi2_array = BinsChiTest(m, pt_value, pt_column, pt_index)

    #### 合并最相近的相邻分箱组
    # 停止条件：unique_intervals.shape[0] <= min_intervals
    # 停止条件：minimum chi2 > threshold 且 the number of unique_intervals <= max_intervals
    while (((chi2_array.min() <= threshold) or (unique_intervals.shape[0] > max_intervals)) and
           (unique_intervals.shape[0] > min_intervals)):
        # 识别具有最小 chi2 分数的相邻对的索引
        index_adjacent_to_merge, = np.where(chi2_array == chi2_array.min())
        # 确定 chi2 分数最小的区间（或区间）
        i_merge = adjacent_index[index_adjacent_to_merge, :]
        # 将每个选定对的区间合并为一个新区间
        new_interval = np.array([(unique_intervals[:, 0][unique_intervals[:, 1] == index[0]][0],
                                  index[1]) for index in i_merge])

        # 删除选定的间隔并添加合并的间隔
        index_delete_merged = np.array(
            [np.where(unique_intervals[:, 1] == e)[0][0] for e in i_merge.reshape(1, -1)[0]])
        unique_intervals = np.vstack((
            np.delete(unique_intervals, index_delete_merged, axis=0), new_interval))
        unique_intervals.sort(axis=0)

        # 使用新的分箱阈值，对值进行重新分配
        intervals, unique_intervals = assign_interval_unique(x, unique_intervals[:, 1])
        pt_value, pt_column, pt_index = pivot_table(intervals[:, 1], y)
        # 对每对 m 个相邻区间执行卡方检验，使用第 i 个区间的索引作为区间对的索引
        adjacent_index, chi2_array = BinsChiTest(m, pt_value, pt_column, pt_index)

    return unique_intervals[:, 1].tolist()


class Chi2Bins(WOEMixin):
    """ chi2 分箱 """
    bins: Optional[Union[int, List[Union[int, float]]]]

    def __init__(
            self,
            variables: Series,
            target: Series,
            init_bins: Optional[int] = 10,
            min_bins: Optional[int] = 3,
            max_bins: Optional[int] = 5,
            m: Optional[int] = 2,
            abnormal_vals: Optional[List[Union[str, int, float]]] = None,
            fill_bin: Optional[bool] = True,
            confidence_level: Optional[float] = 0.9,
            decimal: Optional[int] = 6,
            weight: Optional[Any] = None,
            *args: Any,
            **kwargs: Any,
    ):
        """
        描述：Chi2Bins 分箱

        :param variables: 待分箱变量
        :param target: 目标标签变量
        :param init_bins: 初始化分箱数量，在此分箱数量基础上进行等频分箱，再依据卡方检验进行分箱的合并
        :param min_bins: 最小的分箱数量
        :param max_bins: 最大的分箱数量
        :param m: 目标变量类别数
        :param abnormal_vals: 特殊值分箱，在变量存在特殊值时单独分一箱，如 -1111, -9999。
        :param fill_bin: 在各分箱中偶发性会出现 good 或 bad 为 0 的情况，默认 fill_bin 为 True ，为该分箱填充 0.5。
        :param confidence_level: 卡方检验置信度

        Example:
            woe = Chi2Bins(variables, target)
            woe.fit()
        """
        self.variables = variables
        self.target = target
        self.init_bins = init_bins
        self.min_bins = min_bins
        self.max_bins = max_bins
        self.m = m
        self.abnormal_vals = abnormal_vals
        self.fill_bin = fill_bin
        self.confidence_level = confidence_level
        self.decimal = decimal
        self.args = args
        self.kwargs = kwargs
        self.data_prepare(
            variables=variables, target=target, weight=weight)

    def fit(
            self,
            score: Optional[bool] = True,
            origin_border: Optional[bool] = False,
            order: Optional[bool] = True,
            **kwargs: Any,
    ) -> DataFrame:
        """
        :param score: 是否增加 WOEMixin score。
        :param origin_border: 是否增加 分箱中的最大值与最小值。
        :param order: 是否增加单调性判断。
        :return: DataFrame WOEMixin result.
        """
        self.bins = chi_merge_vector(
            self.variables.to_numpy(), self.target.to_numpy(), m=self.m,
            confidence_level=self.confidence_level, max_intervals=self.max_bins,
            min_intervals=self.min_bins, initial_intervals=self.init_bins,)
        self.bins = self.bins
        self.bins.insert(0, -np.inf)
        _bucket, _bins = pd.cut(
            self.df_data.variables, self.bins, include_lowest=True, retbins=True, duplicates="drop")
        bucket = pd.DataFrame({
            'variables': self.df_data.variables,
            'target': self.df_data.target,
            'bucket': _bucket,
        }).groupby('bucket', as_index=True, observed=False)
        self.cal_woe_iv(bucket, score=score, origin_border=origin_border, order=order)
        return self.woe



from pydantic import BaseModel, Field
from typing import List, Union, Optional


__all__ = [
    "IndexField"
]


class IndexField(BaseModel):
    """"""
    # origin dataset
    db_name: Optional[Union[List[str], str]] = Field(None, description="数据库名")
    table_name: Optional[Union[List[str], str]] = Field(None, description="表名")
    columns: Union[List[str], str] = Field(..., description="数据列名")

    # index name
    index_name: str = Field(..., description="指标英文名")
    index_name_cn: Optional[str] = Field(None, description="指标中文名")

    # date
    start_date: Optional[str] = Field(None, description="依据start_date/end_date进行时间段内的数据衍生")
    end_date: Optional[str] = Field(None, description="依据start_date/end_date进行时间段内的数据衍生")
    load_date: Optional[str] = Field(None, description="依据load_date进行时间点的数据衍生")

    origin_op: Optional[str] = Field(None, description="在衍生之前需要处理的运算")
    operators: Optional[Union[List[str], str]] = Field(..., description="")
    alias_operators: Optional[Union[List[str], str]] = Field(None, description="")
    months: Optional[List[Union[str, int]]] = Field(..., description="衍生月份数")
    id_levels: Union[List[str], str] = Field(..., description="")
    save_db: str = Field(..., description="")
    save_table: str = Field(..., description="")



from typing import Dict, Optional, Callable, Literal


__all__ = ["LoggerMixin"]
TypePrintColor = Literal['black', 'red', 'green', 'yellow', 'blue', 'cyan', 'magenta', 'white', 'reset']


try:
    from colorama import Fore, Style

    print_color_mapping = {
        "BLACK".lower(): Fore.BLACK,
        "RED".lower(): Fore.RED,
        "GREEN".lower(): Fore.GREEN,
        "YELLOW".lower(): Fore.YELLOW,
        "BLUE".lower(): Fore.BLUE,
        "CYAN".lower(): Fore.CYAN,
        "MAGENTA".lower(): Fore.MAGENTA,
        "WHITE".lower(): Fore.WHITE,
        "RESET".lower(): Fore.RESET,
    }


    class LoggerMixin:
        """"""
        logger: Optional[Callable]
        verbose: Optional[bool]

        def _logger(
                self,
                msg: str,
                color: Optional[TypePrintColor] = None
        ) -> None:
            """"""
            if self.logger: self.logger.info(msg)
            if self.verbose:
                if color:
                    print(print_color_mapping.get(color) + msg + Style.RESET_ALL, flush=True)
                else:
                    print(msg, flush=True)

        def _logger_base(self, msg: str, color: Optional[TypePrintColor] = None):
            """"""
            if self.logger: self.logger.info(msg)
            if self.verbose: print(print_color_mapping.get(color) + msg + Style.RESET_ALL, flush=True)

        def _logger_agent_start(
                self,
                msg: Optional[str] = None,
                name: Optional[str] = None,
                color: Optional[TypePrintColor] = "green"
        ):
            """"""
            if msg is None: msg = f"[{name}] Start ...\n"
            self._logger_base(msg=msg, color=color)

        def _logger_agent_end(
                self,
                msg: Optional[str] = None,
                name: Optional[str] = None,
                color: Optional[TypePrintColor] = "green"
        ):
            """"""
            if msg is None: msg = f"[{name}] End ...\n"
            self._logger_base(msg=msg, color=color)

        def _logger_agent_question(
                self,
                msg: Optional[str] = None,
                name: Optional[str] = None,
                content: Optional[str] = None,
                color: Optional[TypePrintColor] = "green"
        ):
            """"""
            if msg is None: msg = f"[{name}] User Question: {content}\n"
            self._logger_base(msg=msg, color=color)

        def _logger_messages_start(
                self,
                msg: Optional[str] = None,
                name: Optional[str] = None,
                color: Optional[TypePrintColor] = "red"
        ):
            """"""
            if msg is None: msg = f"{20 * '='} [{name}] Messages Start {20 * '='}\n"
            self._logger_base(msg=msg, color=color)

        def _logger_messages_end(
                self,
                msg: Optional[str] = None,
                name: Optional[str] = None,
                color: Optional[TypePrintColor] = "red"
        ):
            """"""
            if msg is None: msg = f"{20 * '='} [{name}] Messages End    {20 * '='}\n"
            self._logger_base(msg=msg, color=color)

        def _logger_messages(
                self,
                msg: Optional[str] = None,
                role: Optional[str] = None,
                content: Optional[str] = None,
                color: Optional[TypePrintColor] = "blue"
        ):
            """"""
            if msg is None: msg = f"{role}: [{content}]\n"
            self._logger_base(msg=msg, color=color)

        def _logger_agent_script(
                self,
                msg: Optional[str] = None,
                name: Optional[str] = None,
                script: Optional[str] = None,
                color: Optional[TypePrintColor] = "magenta"
        ):
            """"""
            if msg is None: msg = f"[{name}] Script: ```\n{script}\n```"
            self._logger_base(msg=msg, color=color)

        def _logger_agent_search(
                self,
                msg: Optional[str] = None,
                name: Optional[str] = None,
                content: Optional[str] = None,
                color: Optional[TypePrintColor] = "magenta"
        ):
            """"""
            if msg is None: msg = f"[{name}] Script: ```\n{content}\n```"
            self._logger_base(msg=msg, color=color)

        def _logger_agent_warning(self, msg: str, color: Optional[TypePrintColor] = "red"):
            """"""
            self._logger_base(msg=msg, color=color)

        def _logger_agent_final_answer(
                self,
                msg: Optional[str] = None,
                name: Optional[str] = None,
                content: Optional[str] = None,
                color: Optional[TypePrintColor] = "yellow"
        ):
            """"""
            if msg is None: msg = f"[{name}] Final Answer: {content}\n"
            self._logger_base(msg=msg, color=color)

        def _logger_dict(self, msg: Dict, color="green"):
            """"""
            for key, val in msg.items():
                if self.logger: self.logger.info(f"{key}: {val}")
                if self.verbose: print(print_color_mapping.get(color) + f"{key}: {val}" + Style.RESET_ALL, flush=True)

        def _logger_color(self, msg: str) -> None:
            """"""
            last_line = msg.split('\n')[-1]

            if self.logger: self.logger.info(msg)
            if self.verbose:
                if last_line.startswith("Answer"):
                    print(Fore.YELLOW + msg + Style.RESET_ALL)
                elif last_line.startswith("Action"):
                    print(Fore.GREEN + msg + Style.RESET_ALL)
                elif last_line.startswith("No Action"):
                    print(Fore.RED + msg + Style.RESET_ALL)
                elif last_line.startswith("Observation"):
                    print(Fore.BLUE + msg + Style.RESET_ALL)
                elif last_line.startswith("Running"):
                    print(Fore.CYAN + msg + Style.RESET_ALL)
                else:
                    print(msg)
                print()

except ModuleNotFoundError:
    print(ModuleNotFoundError("pip install colorama"))

    class LoggerMixin:
        """"""
        logger: Optional[Callable]
        verbose: Optional[bool]

        def _logger(
                self,
                msg: str,
                color: Optional[TypePrintColor] = None
        ) -> None:
            """"""
            if self.logger: self.logger.info(msg)
            if self.verbose:
                print(msg, flush=True)





import logging

def upsampling(df, down_rate=None, up_rate=None, target_name='target', label=(0, 1)):
    """

    :param df:
    :param down_rate:
    :param up_rate:
    :param target_name:
    :param label:
    :return:
    """
    logging.warning('未来采样API为sampling, upsampling即将弃用。')
    df_observation = df[df[target_name] == label[1]].copy()
    df_no_obser = df[df[target_name] == label[0]].copy()

    if up_rate:
        num = int(df_no_obser.shape[0] * up_rate)
        df_observation = df_observation.sample(n=num, replace=True)

    if down_rate:
        num = int(df_no_obser.shape[0] * down_rate)
        df_no_obser = df_no_obser.sample(n=num, replace=False)

    df = df_no_obser.append(df_observation)
    return df.sample(frac=1).reset_index(drop=True)


def sampling(df, down_rate=None, up_rate=None, target_name='target', label=(0, 1)):
    """
    上采样与下采样
    :param df:
    :param down_rate:
    :param up_rate:
    :param target_name:
    :param label:
    :return:
    """
    df_observation = df[df[target_name] == label[1]].copy()
    df_no_obser = df[df[target_name] == label[0]].copy()

    if up_rate:
        num = int(df_no_obser.shape[0] * up_rate)
        df_observation = df_observation.sample(n=num, replace=True)

    if down_rate:
        num = int(df_no_obser.shape[0] * down_rate)
        df_no_obser = df_no_obser.sample(n=num, replace=False)

    df = df_no_obser.append(df_observation)
    return df.sample(frac=1).reset_index(drop=True)



import numpy as np
import pandas as pd
from pandas import DataFrame, Series
from typing import Any, List, Union, Optional
from kivi.woe.base import WOEMixin


__all__ = ['KSBins']


class KSBins(WOEMixin):
    """
    描述：决策树分箱
    """
    cutoff_point: List[Union[float, int]]

    def __init__(
            self,
            variables: Series,
            target: Series,
            bins: Optional[int] = 5,
            abnormal_vals: Optional[List[Union[str, int, float]]] = None,
            fill_bin: Optional[bool] = True,
            decimal: Optional[int] = 6,
            weight: Optional[Any] = None,
            min_sample_rate: Optional[float] = 0.05,
            *args: Any,
            **kwargs: Any,
    ):
        """
        描述：最优KS分箱方法，使用决策树最优 entropy 进行分箱分析。

        :param variables: 待分箱变量
        :param target: 目标标签变量
        :param bins: 决策树分箱中最大的叶子结点数量，一般对应的是最终分箱数量，默认为 5 。
        :param abnormal_vals: 特殊值分箱，在变量存在特殊值时单独分一箱，如 -1111, -9999。
        :param fill_bin: 在各分箱中偶发性会出现 good 或 bad 为 0 的情况，默认 fill_pos 为 True ，为该分箱填充 0.5。

        Example:
            woe = KSBins(variables, target, bins=5, fill_bin=True)
            woe.fit()
        """
        self.variables = variables
        self.target = target
        self.bins = bins
        self.fill_bin = fill_bin
        self.abnormal_vals = abnormal_vals
        self.decimal = decimal
        self.args = args
        self.kwargs = kwargs
        self.data_prepare(
            variables=variables, target=target, weight=weight)
        if min_sample_rate is not None:
            self.min_sample = min_sample_rate * len(self.target)

    def max_ks_value(self, data):
        """
        """
        bad = data.target.sum()
        good = data.target.count() - bad

        data_cum = pd.crosstab(data.variables, data.target)\
            .div([good, bad]).cumsum()
        ks_list = (data_cum[1] - data_cum[0]).abs()
        ks_list_index = ks_list.nlargest(len(ks_list)).index.tolist()

        for i in ks_list_index:
            group_a = data.target[data.variables <= i].count()
            group_b = data.target[data.variables > i].count()
            if group_a >= self.min_sample and group_b >= self.min_sample:
                return i

    def get_cutoff_point(self, ):
        """"""
        bins = [-np.inf, np.inf]
        data_upgrade = self.df_data.copy()
        for i in range(self.bins - 1):
            bins.append(self.max_ks_value(data_upgrade))
            bins.sort()

            if len(bins) <= self.bins:
                # upgrade max sample bin
                df_buckets_stats = pd.cut(
                    self.df_data.variables, bins, include_lowest=True, duplicates='drop').value_counts(sort=True)
                max_bin = df_buckets_stats.idxmax()
                max_bin = [max_bin.left, max_bin.right]

                data_upgrade = self.df_data[self.df_data.variables.between(*max_bin, inclusive='right')].copy()
        self.cutoff_point = bins

    def fit(
            self,
            score: Optional[bool] = True,
            origin_border: Optional[bool] = False,
            order: Optional[bool] = True,
            **kwargs: Any,
    ) -> DataFrame:
        """

        :param score: 是否增加 WOEMixin score。
        :param origin_border: 是否增加 分箱中的最大值与最小值。
        :param order: 是否增加单调性判断。
        :return: DataFrame WOEMixin result.
        """
        self.get_cutoff_point()
        _bucket, _bins = pd.cut(self.df_data.variables, self.cutoff_point, include_lowest=True, retbins=True)
        bucket = pd.DataFrame({
            'variables': self.df_data.variables,
            'target': self.df_data.target,
            'bucket': _bucket,
        }).groupby('bucket', as_index=True, observed=False)
        self.cal_woe_iv(bucket, score=score, origin_border=origin_border, order=order)
        return self.woe





from .Linear import LinearRegression
from .Lasso import Lasso
from .Ridge import Ridge
from .Tree import Tree
from .LogisticRegression import Logistic
from .Apriori import Apriori
from .simtext import similarity



import pandas as pd
from kivi.evaluate.utils import BinsMixin


__all__ = ["CSI"]


class CSI(BinsMixin):
    def __init__(self, score=None):
        """
        Des: 计算CSI
        :param score: 分箱分数
        :param margin: cutoff point 左右填充预留边界
        """
        super(CSI, self).__init__()
        self.score = score

    def _csi(self, res):
        """
        Des: 计算CSI的核心函数
        :param res: 明细表
        :return: 明细表以及CSI值 res, csi
        """

        res['Expected_percentage'] = res.Expected / res.Expected.sum()
        res['Actual_percentage'] = res.Actual / res.Actual.sum()
        res['A-E'] = res.Actual_percentage - res.Expected_percentage
        if len(res) != len(self.score):
            raise Exception(TypeError, 'Score len error!')
        res['score'] = self.score
        res['index'] = res['A-E'] * self.score
        csi = res['index'].sum()
        return res, csi

    def Continue(self, expected, actual, bins=10, binType='distance', cutPoint=None):
        """
        Des: 连续型变量的CSI指标
        :param expected: 预期分布
        :param actual: 实际分布
        :param bins: 分箱数目
        :param binType: 分箱类型 'distance'等距 or 'frequency'等频 or 'customize'自定义
        :param csvName: 保存CSV Name, 默认不保存
        :return: 明细表以及CSI值 res, CSI
        """
        expected.dropna(inplace=True)
        actual.dropna(inplace=True)

        expectedMax = expected.max()
        expectedMin = expected.min()
        actualMax = actual.max()
        actualMin = actual.min()

        if not pd.api.types.is_numeric_dtype(expected):
            raise Exception(TypeError, 'expected must be numeric!')

        if not pd.api.types.is_numeric_dtype(actual):
            raise Exception(TypeError, 'actual must be numeric!')

        if (expectedMax == expectedMin) or (actualMax == actualMin):
            raise Exception()

        if (expectedMin > actualMin) or (expectedMax < actualMax):
            raise Exception()

        if binType == 'distance':
            cutoffPoint = self._cutoffPoint(
                bins=bins, min=expectedMin, max=expectedMax)
        elif binType == 'frequency':
            cutoffPoint = self._cutoffPoint(
                bins=bins, percent=True, expected=expected)
        elif binType == 'customize' and cutPoint:
            cutoffPoint = cutPoint
        else:
            raise Exception(TypeError, '分箱错误!')

        res = pd.DataFrame({
            'Expected': pd.cut(expected, cutoffPoint).value_counts(sort=False),
            'Actual': pd.cut(actual, cutoffPoint).value_counts(sort=False),
        },)

        return self._csi(res)

    def Category(self, expected, actual):
        """
        Des: 类别型变量的CSI
        :param expected:
        :param actual:
        :return: 明细表以及CSI值 res, CSI
        """
        if (not pd.api.types.is_string_dtype(expected)) or (not pd.api.types.is_object_dtype(expected)):
            raise Exception(TypeError, 'expected must be string or object!')
        if (not pd.api.types.is_string_dtype(actual)) or (not pd.api.types.is_object_dtype(actual)):
            raise Exception(TypeError, 'actual must be string or object!')

        expected.dropna(inplace=True)
        actual.dropna(inplace=True)

        expectedGroup = expected.value_counts(sort=False)
        actualGroup = actual.value_counts(sort=False)

        res = pd.merge(
            pd.DataFrame({'category': expectedGroup.index, 'Expected': expectedGroup.values}),
            pd.DataFrame({'category': actualGroup.index, 'Actual': actualGroup.values}),
            on='category',
            how='outer',)

        self.res, self.csi = self._csi(res)
        return self.res, self.csi



import numpy as np
import pandas as pd
from typing import Literal


__all__ = [
    "int_score",
    "detect_monotony",
    "monotony",
]


TypeOrder = Literal["上升下降", "下降上升", "单调上升", "单调下降", "数据不足", "未知"]


def int_score(score):
    """
    将数值分数按 5 取整
    :param score: 浮点型分数
    :return: 按 5 取整分数
    """
    basic_score = np.arange(0, 101, 5)
    return basic_score[np.argmin(np.abs(score - basic_score))]


def detect_monotony(trend_mask):
    """"""
    if all(trend_mask):
        return '单调上升'
    elif not any(trend_mask):
        return '单调下降'
    else:
        return None


def monotony(values: pd.Series):
    """"""
    diff = values.diff().dropna()
    if len(diff) > 0:
        trend_mask = (diff >= 0).tolist()
        trend = detect_monotony(trend_mask)
        if trend:
            return trend
        else:
            for mask_id in range(len(trend_mask)):
                top_trend = detect_monotony(trend_mask[: mask_id])
                end_trend = detect_monotony(trend_mask[mask_id:])
                if all([top_trend, end_trend]):
                    trend = f'{top_trend[2:]}{end_trend[2:]}'
                    return trend
            return '未知'
    else:
        return '数据不足'




import sys
import numpy as np
import pandas as pd
from pandas import DataFrame
from typing import Any, Dict, List, Tuple, Callable, Optional
from ..utils.utils import dispatch_tqdm
from ..utils.logger import LoggerMixin
from .woe_obj import *


__all__ = [
    "ManualBinsTool",
    "WOEBatch",
]


class ManualBinsTool(LoggerMixin):
    """
    """
    def __init__(
            self,
            df: Optional[DataFrame] = None,
            target_name: Optional[str] = 'target',
            abnormal_vals: Optional[List[Union[str, int, float]]] = None,
            logger: Optional[Callable] = None,
            verbose: Optional[bool] = True,
            *args: Any,
            **kwargs: Any
    ):
        """
        :param df: 数据集
        :param target_name: 目标变量名
        :param abnormal_vals: 异常值
        :param logger: 日志
        :param verbose: 是否打印日志
        """
        self.df = df
        self.target_name = target_name
        self.abnormal_vals = abnormal_vals
        self.logger = logger
        self.verbose = verbose
        self.args = args
        self.kwargs = kwargs

    @staticmethod
    def append_rebin_woe(df_woe: DataFrame, df_rebin: DataFrame) -> DataFrame:
        """
        为原始分箱增加其他类型分箱
        :param df_woe: 原始分箱
        :param df_rebin: 手工分箱
        """
        var_names = df_rebin.var_name.unique().tolist()
        df_origin = df_woe[df_woe.var_name.isin(var_names)].copy()
        
        origin_columns = df_origin.columns.tolist()
        rebin_columns = df_rebin.columns.tolist()
        subtract_columns = list(set(origin_columns) - set(rebin_columns)) + ['var_name']
    
        df_rebin = df_rebin.merge(df_origin[subtract_columns].drop_duplicates(), on='var_name')
        df_rebin = df_rebin[origin_columns]
    
        df_woe = df_woe[~df_woe.var_name.isin(var_names)]
        df_woe = pd.concat([df_woe, df_rebin], axis=0, ignore_index=True)
        return df_woe
    
    def manual_rebin(
            self,
            column: str,
            bins: List[Union[float, int]],
    ) -> DataFrame:
        """ 描述：指定分箱阈值进行重分箱。 """
        bins = ManuallyBins(
            variables=self.df[column], target=self.df[self.target_name],
            bins=bins, fill_bin=True, abnormal_vals=self.abnormal_vals)
        df_rebin_woe = bins.fit()
        ipython_woe = df_rebin_woe.style.bar(
            subset=['bad_rate', 'woe', 'score'],
            align='mid', color=['#d65f5f', '#5fba7d'],)
        if self.verbose:
            if 'ipykernel' in sys.modules:
                from IPython.display import display
                display(ipython_woe)
            else:
                print(df_rebin_woe)
        return df_rebin_woe

    def _add_nan_row_to_woe(
            self,
            woe: DataFrame,
            score: Optional[int] = None,
            fill_score: Optional[Dict[str, int]] = None,
    ) -> DataFrame:
        """"""
        nan_row = woe.tail(1).copy()
        order = nan_row["order"].tolist()[0]
        for column in ["min_bin", "max_bin"]:
            nan_row[column] = np.nan

        for column in ["total", "bad", "bad_rate", "woe", "iv"]:
            nan_row[column] = 0.

        if score is not None:
            nan_row["score"] = score
        elif fill_score is not None and fill_score.get(order):
            nan_row["score"] = fill_score.get(order)
        else:
            raise ValueError(f"score is None or {order} not find in fill_score.")

        woe = pd.concat([woe, nan_row], axis=0, ignore_index=True)
        woe.reset_index(drop=True, inplace=True)
        return woe

    def add_nan_bin(
            self,
            df_woe: Optional[DataFrame] = None,
            columns: Optional[Union[List[str], Dict[str, int]]] = None,
            fill_score: Optional[Dict[str, int]] = None,
    ) -> DataFrame:
        """
        :param df_woe: woe分箱结果
        :param columns: 需要填充的变量名，有以下两种格式可以选择
            - `List[str]`: 如` ["age", "education"] `是对这两个字段进行空箱默认值的填充，默认值参考参数`fill_score`
            - `Dict[str, int]`: 如 ` {"age": 15, "duration": 30} ` 会对 `age` 自动空箱填充为15分，`duration` 字段填充为30分
        :param fill_score: 填充的分值，默认如下
            fill_score = {
                "单调上升": 95, "单调下降": 5,
                "上升下降": 30, "下降上升": 30,
                "数据不足": 30, "未知": 30
            }

        1. 对全部分箱的空值进行填充
        2. 对指定的变量分箱进行填充
        3. 对指定的变量填充指定的分值
        """
        if fill_score is None:
            fill_score = {
                "单调上升": 95, "单调下降": 5,
                "上升下降": 30, "下降上升": 30,
                "数据不足": 30, "未知": 30
            }

        nan_woe_list = []
        if columns is None:
            columns = df_woe.var_name.unique().tolist()

        if isinstance(columns, list):
            for column in columns:
                df_nan_woe = df_woe[df_woe.var_name == column].copy()
                if df_nan_woe.min_bin.isna().sum() == 0:
                    df_nan_woe = self._add_nan_row_to_woe(df_nan_woe, fill_score=fill_score)
                    nan_woe_list.append(df_nan_woe)
        elif isinstance(columns, dict):
            for column, score in columns.items():
                df_nan_woe = df_woe[df_woe.var_name == column].copy()
                if df_nan_woe.min_bin.isna().sum() == 0:
                    df_nan_woe = self._add_nan_row_to_woe(df_nan_woe, score=score)
                    nan_woe_list.append(df_nan_woe)

        df_nan_woe = pd.concat(nan_woe_list, axis=0, ignore_index=True)
        df_woe = self.append_rebin_woe(df_woe, df_rebin=df_nan_woe)
        return df_woe


class WOEBatch(LoggerMixin):
    """
    todo: 自动判断分箱的数据类型，对类别型变量选择合适的分箱方式。需要检测类别的数目。
    todo: 按照月份进行跑批，规避不同月份中出现同一UUID的情况
    """
    def __init__(
            self,
            df: DataFrame,
            bins: Optional[Union[int, Dict[str, List[Union[float, int]]]]] = 5,
            max_bin: Optional[int] = 6,
            min_bin: Optional[int] = 3,
            rebin: Optional[bool] = True,
            woe_type: TypeWoe = TreeBins,
            woe_columns: Optional[List[str]] = None,
            target_name: Optional[str] = "target",
            observe_order: Optional[List[str]] = None,
            abnormal_vals: Optional[List[Union[str, int, float]]] = None,
            protect_columns: Optional[List[str]] = None,
            logger: Optional[Callable] = None,
            verbose: Optional[bool] = False,
            *args: Any,
            **kwargs: Any,
    ):
        """
        描述：批量计算WOE IV值

        参数：
        :param df: DataFrame 原始数据集。
        :param bins[int, dict]: [int]分箱数量，默认为 5；[dict(var_name: bins)]依据截断点分箱。
        :param max_bin: 允许最大分箱数量。
        :param min_bin: 允许最小分箱数量。
        :param rebin: 是否重新分箱，默认为 `True`。
        :param woe_type: 分箱方法，默认为决策树分箱 `TreeBins`。
        :param woe_columns: 需要进行分箱的字段，默认为全部字段。
        :param target_name: 标签名称，默认为 `target`。
        :param observe_order: 保留分箱单调性，默认为 `['单调上升','单调下降']`；即对不满足条件的数据进行重分箱。
        :param abnormal_vals: 需要剔除异常值，如[-9099, -4404]
        :param protect_columns: 需要剔除分箱的字段, 如不需要进行分箱的字段['uuid', 'target']。
        :param logger: 日志记录器，默认为 `None`。
        :param verbose: 是否打印分箱日志，默认为 `False`。

        Example：
            df_woe = WOEBatch(df, rebin=False)
            df_woe = WOEBatch(df, max_bins=5, min_bins=2, rebin=True)
        """
        self.df = df
        self.bins = bins
        self.max_bin = max_bin
        self.min_bin = min_bin
        self.woe_type = woe_type
        self.woe_columns = woe_columns
        self.target_name = target_name
        self.abnormal_vals = abnormal_vals
        self.rebin = rebin
        self.logger = logger
        self.verbose = verbose
        self.args = args
        self.kwargs = kwargs

        self.woe_container = []
        self.error_woe_columns = []
        if protect_columns is None:
            self.protect_columns = ["uuid", "target"]
        else:
            self.protect_columns = protect_columns
        if observe_order is None:
            self.observe_order = ['单调上升', '单调下降']
        self._get_woe_columns()

    def __call__(self, *args, **kwargs):
        """"""

    def _get_woe_columns(self):
        """"""
        total_columns = self.df.columns.tolist()
        if self.woe_columns is None:
            for column in self.protect_columns:
                if column in total_columns:
                    total_columns.remove(column)
            self.woe_columns = total_columns

    def _detect_observe_bins(self, df_woe: DataFrame) -> Tuple[DataFrame, DataFrame, List[str]]:
        """"""
        df_woe_observe = df_woe[df_woe.order.isin(self.observe_order)].copy()
        df_woe_rebin = df_woe[~df_woe.order.isin(self.observe_order)].copy()
        rebin_columns = df_woe_rebin.var_name.unique().tolist()
        return df_woe_observe, df_woe_rebin, rebin_columns

    def woe_batch(self,) -> DataFrame:
        """"""
        self._logger(msg=f"[{__class__.__name__}] WOE Batch samples shape: {self.df.shape}.\n", color="green")
        for woe_column in dispatch_tqdm(self.woe_columns, desc=f'[{__class__.__name__}] Batch WOE'):
            try:
                if isinstance(self.woe_type, ManuallyBins):
                    _bins = self.bins.get(woe_column)
                else:
                    _bins = self.bins
                woe_fun = self.woe_type(
                    variables=self.df[woe_column],
                    target=self.df[self.target_name],
                    abnormal_vals=self.abnormal_vals,
                    bins=_bins,
                    **self.kwargs
                )
                woe = woe_fun.fit()
                self.woe_container.append(woe)
            except Exception as e:
                self._logger(msg=f"[{__class__.__name__}] {woe_column} WOE Error: {e}", color="red")
                self.error_woe_columns.append({woe_column: e})
        df_woe = pd.concat(self.woe_container, axis=0, ignore_index=True)
        return df_woe

    def _set_woe_batch(self, woe_columns: Optional[List[str]] = None) -> DataFrame:
        """"""
        if woe_columns:
            self.woe_columns = woe_columns
        self.bins = self.max_bin
        self.woe_container = []
        df_woe = self.woe_batch()
        return df_woe

    def woe_batch_with_rebin(self, ) -> DataFrame:
        """"""
        df_woe = self._set_woe_batch()
        df_woe_observe, df_woe_rebin, rebin_columns = self._detect_observe_bins(df_woe)

        self._logger(msg=f'[{__class__.__name__}] Bins: {self.max_bin}; Success: {len(df_woe_observe)}; Rebin: {len(df_woe_rebin)}; Error: {len(self.error_woe_columns)}', color='green')
        df_woe_total = df_woe_observe.copy()

        self.max_bin -= 1
        while len(rebin_columns) > 0 and self.max_bin >= self.min_bin:
            df_woe = self._set_woe_batch(woe_columns=rebin_columns)
            df_woe_observe, df_woe_rebin, rebin_columns = self._detect_observe_bins(df_woe)
            df_woe_total = pd.concat([df_woe_total, df_woe_observe], axis=0, ignore_index=True)
            self._logger(
                msg=f'[{__class__.__name__}] Bins: {self.max_bin}; Success: {len(df_woe_observe)}; Rebin: {len(df_woe_rebin)}; Error: {len(self.error_woe_columns)}', color='green')
            self.max_bin -= 1
        df_woe_total = pd.concat([df_woe_total, df_woe_rebin], axis=0, ignore_index=True)
        return df_woe_total

    def fit(self) -> DataFrame:
        """"""
        if self.rebin:
            df_woe = self.woe_batch_with_rebin()
        else:
            df_woe = self.woe_batch()
        return df_woe



import numpy as np
import pandas as pd
import logging
from .utils import BetaDistribution


def get_omega(
        now_y: pd.Series = None,
        previous_y: pd.Series = None,
        tau=None,
):
    """
    获取ODD校准的omega值，给定采样前后的样本，或者给定tau概率
    :param now_y:
    :param previous_y:
    :param tau:
    :return:
    """
    if previous_y is not None:
        tau = sum(previous_y) / len(previous_y)
    if now_y is not None:
        y_ratio = sum(now_y) / len(now_y)
    omega = (tau / (1-tau)) * ((1 - y_ratio) / y_ratio)
    return omega


def calibration(X, coef, intercept, omega):
    """
    概率校准
    :param X: DataSet
    :param coef: 截距
    :param intercept: 斜率
    :param omega: omega 参数
    :return: 校准后的pd
    """
    calibration_pd = 1 / (1 + np.e ** (- intercept - np.log(omega) - np.dot(X, coef)))
    return calibration_pd

class BetaCalibration(BetaDistribution):
    """
    基本步骤:
        1. 确定各个评级占比及对应的得分区间；
        2. 统计落入各个评级的客户数及违约数；
        3. 二项检验；
        4. 计算校准后的实际违约率。
    """
    def __init__(
            self,
            Alpha: float,
            Beta: float,
            target_grade: int,
            alpha: float = 0.05):
        super(BetaCalibration, self).__init__(Alpha, Beta)
        self.alpha = alpha
        self.target_grade = target_grade

        self.target_grade_rate, self.CDF = self.grade_rate(self.target_grade)
        self.arange = np.arange(1, target_grade+1, 1)
        self.calibration_distritution = pd.DataFrame({
            'rating': self.arange,
            'percentile': [f'{i}/{target_grade}'for i in self.arange],
            'accum_dis': self.target_grade_rate,
            'distribution': self.CDF,
        })


    def score_to_rank(self, predict_default=None, score=None):
        """
        得分映射区间
        :return:
        """
        if predict_default is not None:
            # 计算 Beta 分布下分位点的 PD 值
            self.quantile = [np.quantile(predict_default, cdf) for cdf in self.CDF]
        if score is not None:
            # 计算 Beta 分布下分位点的 Score 值
            self.quantile = [np.quantile(score, cdf) for cdf in self.CDF]
        self.quantile.insert(0, 0)
        if len(np.unique(self.quantile)) != len(self.quantile):
            logging.error('等级过细，存在相同数值截断点！')

    def count_default(self, target, predict_default=None, score=None, right=True):
        """

        :param target:
        :param predict_default:
        :param score:
        :param right:
        :return:
        """
        if predict_default is not None:
            pass
        if score is not None:

            buckets = pd.DataFrame({
                'score': score,
                'target': target,
                'bucket': pd.cut(score, self.quantile, include_lowest=True, right=right, duplicates='drop')
            }).groupby('bucket', as_index=True)

            # 校准后分布
            self.real_data_distribution = pd.DataFrame({
                'rating': self.arange,
                'bin_samples': buckets.target.count(),
                'default_count': buckets.target.sum()
            })

    def to_rank(self, predict_default: float, grade: list):
        """
        PD 到主标尺等级的映射
        :param pd: 预测PD
        :param ranking: 主标尺分箱的max bin
        :return: 等级
        """
        return sum(np.array(grade) <= predict_default) - 1

    def binomial_test(self, ):
        # ranking = [ for ]
        # bt = BinomialTest(ranking=, target=, proba=)
        # bt.binomial_boundary()
        # return
        return None


if __name__ == '__main__':
    np.random.seed(1)
    ranking = np.random.randint(0, 100, size=100)
    target = np.random.binomial(1, 0.3, size=100)
    bc = BetaCalibration(2.65, 2.95, 19, ranking, target, proba)
    print(bc.target_grade_rate, bc.CDF)
    # 根据 score predict_default 找截断点
    print(bc.score_to_rank(score=ranking))
    bc.count_default(target=target, score=ranking)



from pydantic import BaseModel, Field
from typing import List, Dict, Union, Optional, Sequence


__all__ = [
    "FeatureEvaluateSchema",
    "StepwiseReport",
]


class FeatureEvaluateSchema(BaseModel):
    """"""
    var_name: str = Field(default=None, description="")
    bad_rate: Optional[float] = Field(default=None, description="")
    desc: Optional[str] = Field(default=None, description="")
    missing: Optional[float] = Field(default=None, description="")

    miss_bad_rate: Optional[float] = Field(default=None, description="")
    not_miss_bad_rate: Optional[float] = Field(default=None, description="")
    null_lift: Optional[float] = Field(default=None, description="")
    not_null_lift: Optional[float] = Field(default=None, description="")

    min: Optional[float] = Field(default=None, description="")
    max: Optional[float] = Field(default=None, description="")
    std: Optional[float] = Field(default=None, description="")
    mode: Optional[float] = Field(default=None, description="")

    skew: Optional[float] = Field(default=None, description="")
    kurt: Optional[float] = Field(default=None, description="")
    cv: Optional[float] = Field(default=None, description="")
    unique: Optional[float] = Field(default=None, description="")

    bad: Optional[Union[int, float]] = Field(default=None, description="")
    count: Optional[Union[int, float]] = Field(default=None, description="")
    good: Optional[Union[int, float]] = Field(default=None, description="")
    ks_test: Optional[float] = Field(default=None, description="")

    chi_statistic: Optional[float] = Field(default=None, description="")
    chi_pvalue: Optional[float] = Field(default=None, description="")

    R2: Optional[float] = Field(default=None, description="")
    intercept: Optional[float] = Field(default=None, description="")
    pvalue_intercept: Optional[float] = Field(default=None, description="")
    param: Optional[float] = Field(default=None, description="")
    pvalue_param: Optional[float] = Field(default=None, description="")

    auc: Optional[float] = Field(default=None, description="")
    ks: Optional[float] = Field(default=None, description="")
    fpr: Optional[Sequence] = Field(default=None, description="")
    tpr: Optional[Sequence] = Field(default=None, description="")


class StepwiseReport(BaseModel):
    """Stepwise report schema"""

    feature: List[str] = Field(default=None, description="")
    auc: Optional[float] = Field(default=None, description="")
    ks: Optional[float] = Field(default=None, description="")
    indicator_change: Optional[float] = Field(default=None, description="")
    num_features: Optional[int] = Field(default=None, description="")
    params: Optional[Dict[str, float]] = Field(default=None, description="")
    woe_impact: Optional[List[str]] = Field(default=None, description="")
    positive_impact: Optional[List[str]] = Field(default=None, description="")
    negative_impact: Optional[List[str]] = Field(default=None, description="")



from pyspark.sql import functions as F


def ModelResultSpark(df_match, score_shceme='sigmod'):
    """
    描述：使用spark计算模块结果，计算模型结果

    参数：
    :param df_match: 匹配分数表
    :param score_shceme: 分数计算方案 `simode`, `linear`
    :return pyspark DataFrame 模型结果表

    示例：
    >>>
    """
    if score_shceme == 'simode':
        weight_type = 'coef'
    elif score_shceme == 'linear':
        weight_type = 'wight'

    df_module = df_match.groupBy('load_date', 'uuid', 'module_name').agg(
        (-(F.sum(F.col('woe') * F.col(f'var_{weight_type}')) + F.first('var_intercept'))).alias('module_ln_odds'),
        (1 / (1 + F.exp(-(F.sum(F.col('woe') * F.col(f'var_{weight_type}')) + F.first('var_intercept'))))).alias(
            'module_pd'),
        F.first('k').alias('k'),
        F.first(f'module_{weight_type}').alias(f'module_{weight_type}'),
        F.first('module_intercept').alias('module_intercept'),
        F.first('module_min_pd').alias('module_min_pd'),
        F.first('module_max_pd').alias('module_max_pd'), )

    df_module = df_module.withColumn('sig', F.when(F.col('module_coef') > 0, 1).otherwise(-1)) \
        .withColumn('module_score', ((F.col('sig') * F.lit(100) / (F.col('module_max_pd') - F.col('module_min_pd'))) * (
                F.col('module_pd') - F.col('module_min_pd'))))

    df_model = df_module.groupBy('load_date', 'uuid').agg(
        (-(F.sum(F.col('module_score') * F.col(f'module_{weight_type}')) + F.first('module_intercept'))).alias(
            'modle_ln_odds'),
        (1 / (1 + F.first('k') * F.exp(
            -(F.sum(F.col('module_score') * F.col(f'module_{weight_type}')) + F.first('module_intercept'))))).alias(
            'modle_pd'),
    )
    df_model_result = df_module.join(df_model, on=['load_date', 'uuid'], how='left')

    return df_model_result



import os, re
import logging
try:
    from pyspark.sql import SparkSession
    from pyspark.sql import functions as F
    from pyspark.sql.types import StructType, StructField, StringType
    from pyspark.sql.window import Window
    from pyspark.sql import DataFrame as TypeSparkDataFrame
except ImportError:
    print("pyspark not installed, please install it first. < pip install pyspark >")
from typing import List, Optional
from tqdm import tqdm
import pandas as pd
from pandas import DataFrame as TypePandasDataFrame
from ..utils.utils import mkdir
from .column import get_table_columns

__all__ = [
    "drop_table",
    "rename_table",
    "find_tables",
    "table_exit",
    "create_table_by_schema",
    "spark_frame_to_hive",
    "insert_into_hive",
    "get_db_table",
    "hive_table_to_csv",
    "save_pandas_frame_to_hive",
]


def drop_table(db_name: str, table_name: str, spark: SparkSession) -> None:
    """
    描述： 给的库名、表名，删除相应的Table

    参数：
    :param db_name: 库名
    :param table_name: 表名，接受 str 或 list
    :param spark:
    :return None:
    """

    def drop(db_name, table_name, spark=None):
        if spark.catalog._jcatalog.tableExists(f'{db_name}.{table_name}'):
            print(f'drop old table {db_name}.{table_name}')
            spark.sql(f'drop table if exists {db_name}.{table_name}')
            print(f'Table {db_name}.{table_name} has been droped!')
        else:
            print(f'Table {db_name}.{table_name} not exit!')

    if isinstance(table_name, list):
        for name in table_name:
            drop(db_name, name, spark)
    elif isinstance(table_name, str):
        drop(db_name, table_name, spark)
    else:
        raise TypeError('table_name')


def rename_table(
        pre_db_name: str,
        pre_table_name: str,
        db_name: str,
        table_name: str,
        spark: SparkSession
) -> None:
    print(f'Pre table {pre_db_name}.{pre_table_name}')
    if db_name is None:
        db_name = pre_db_name
    if table_name:
        spark.sql(f'ALTER TABLE {pre_db_name}.{pre_table_name} RENAME TO {db_name}.{table_name}')
    print(f'Rename to {db_name}.{table_name}')


def find_tables(
        db_name: str,
        tables: str,
        pat: str,
        spark: SparkSession
) -> List[str]:
    """
    描述：选择相应模式的Table

    参数：
    :param db_name: 库名。db_name与tables二选一指定
    :param tables: 表名。db_name与tables二选一指定
    :param pat: 表名正则模式。
    :param spark: spark
    :return : table list for pat.
    """
    if db_name is None and tables is None:
        raise ValueError("db_name, tables")

    elif db_name is not None and tables is None:
        db_tables = spark.catalog.listTables(db_name)
        tables = [table_info.name for table_info in db_tables]

    pattern = re.compile(pat)
    find_res = list(map(pattern.findall, tables))
    find_tables = []
    for table, mask_info in zip(tables, find_res):
        if len(mask_info) != 0:
            find_tables.append(table)
    return find_tables


def table_exit(db_name: str, table_name: str, spark: SparkSession) -> bool:
    """"""
    if spark.catalog._jcatalog.tableExists(f'{db_name}.{table_name}'):
        return True
    else:
        return False


def create_table_by_schema(
        schema: StructType,
        db_name: str,
        table_name: str,
        partition_by: str,
        spark: SparkSession,
        mode='overwrite'
) -> None:
    """
    描述：依据schema创建表

    Example:
        schema = StructType([
            StructField('uuid', StringType(), False),
            StructField('biz_code', StringType(), False),
            StructField('index_val', StringType(), True),
            StructField('load_date', StringType(), False),
        ])
    """
    df = spark.createDataFrame(spark.sparkContext.emptyRDD(), schema=schema)
    df.write.mode(mode).partitionBy(partition_by).saveAsTable(f'{db_name}.{table_name}')


def spark_frame_to_hive(
        df: TypeSparkDataFrame,
        db_name: str,
        table_name: str,
        spark: SparkSession,
        comment: Optional[str] = None,
        partition_by: Optional[str] = None,
        mode: str = 'overwrite',
        repartition: Optional[int] = None,
        verbose: Optional[bool] = True,
) -> None:
    """
    原来的 FrameToHive
    保存表格到数据库
    """
    if verbose: logging.warning(f'Save DataFrame To {db_name}.{table_name}!')

    if partition_by:
        df.write.mode(mode).partitionBy(partition_by).saveAsTable(f'{db_name}.{table_name}')
    elif repartition:
        df.repartition(repartition).write.mode(mode).saveAsTable(f'{db_name}.{table_name}')
    elif spark:
        drop_table(db_name, table_name, spark)
        df.createOrReplaceTempView(f'{table_name}')
        spark.sql(f'create table {db_name}.{table_name} as select * from {table_name}')
        if comment is not None:
            spark.sql(f"ALTER TABLE {db_name}.{table_name} SET TBLPROPERTIES('comment'='{comment}')")
    else:
        raise Exception(f'table {db_name}.{table_name} save failed!')

    if verbose: logging.warning(f'Save End!')


def insert_into_hive(
        df: TypeSparkDataFrame,
        db_name: str,
        table_name: str,
        partition: List[str],
        partition_val: List[str],
        spark: SparkSession ,
        query_mode: Optional[bool] = True
) -> None:
    """
    描述：向Hive表中插入数据

    Example:
        InsertIntoHive(
            df_2, 'tmp_work', 't_cch_uuid_info',
            partition=['biz_code', 'load_date'],
            partition_val=['industry', '20190101'], spark=spark)
    """
    if query_mode:

        part_conditions = []
        for part, part_val in zip(partition, partition_val):
            part_conditions.append(f"{part}='{part_val}'")
        if len(part_conditions) == 1:
            part_condition = str(f'{part_conditions[0]}')
        else:
            part_condition = str(', '.join(part_conditions))

        columns = get_table_columns(db_name=db_name, table_name=table_name, spark=spark)
        for column in partition:
            if column in columns:
                columns.remove(column)

        df.createOrReplaceTempView("temp")
        df_empty = spark.sql(f"""
        INSERT OVERWRITE TABLE {db_name}.{table_name}
        PARTITION ({part_condition}) 
        SELECT {', '.join(columns)}
        FROM temp
        """)
    else:
        df.write.insertInto(f'{db_name}.{table_name}', True)


def get_db_table(db_name: str, spark: SparkSession) -> TypePandasDataFrame:
    """
    desc: get database table info.
    :param db_name: database name
    :param spark: spark session
    :return: pd.DataFrame, table info.
    """
    tables = spark.catalog.listTables(db_name)
    list_table = []
    for table in tables:
        table_info = {
            'name': table.name,
            'database': table.database,
            'description': table.description,
            'tableType': table.tableType,
            'isTemporary': table.isTemporary,
        }
        list_table.append(table_info)
    df_tables = pd.DataFrame(list_table)
    return df_tables


def hive_table_to_csv(csv_path: str, db_name: str = None, table_name: str = None, query: str = None):
    """
    描述：提取Hive Table至CSV文件

    参数：
    :param db_name: 数据库名
    :param table_name: 表名
    :param csv_path: CSV路径
    :param query: 需要导出的 SQL Query

    Example:
        query = 'SELECT * FROM tmp_work.t_cch_sample_for_yto'
        csv_path = './dataset/samples.csv'
        HiveTableToCSV(csv_path=csv_path, query=None)
        HiveTableToCSV(db_name='tmp_work', table_name='t_cch_sample_for_yto', csv_path=csv_path)
    """
    mkdir(csv_path)
    if query:
        shell = f"hive -e 'set hive.cli.print.header=true; {query}' > '{csv_path}'"
    else:
        shell = f"hive -e 'set hive.cli.print.header=true; SELECT * FROM {db_name}.{table_name}' > '{csv_path}'"
    info = os.popen(shell).readlines()
    print('\n'.join(info))


def save_pandas_frame_to_hive(
        df: TypePandasDataFrame,
        db_name: str,
        table_name: str,
        reset: Optional[bool] = True,
        spark: Optional[SparkSession] = None,
        batch: Optional[int] = 5e4,
        verbose: Optional[bool] = False
) -> None:
    """
    描述：将模型结果保存至Hive表中，以5万为默认保存批次
    Example:
        SaveBigDFToHive(df_now_woe, 'tmp_work', 't_cch_sample_woe_20210531', spark)
    """
    table_exist = table_exit(db_name=db_name, table_name=table_name, spark=spark)

    if reset and table_exist:
        drop_table(db_name, table_name, spark)

    for i in tqdm(range(int(len(df) // batch) + 1)):
        low = int(i * batch)
        up = int((i + 1) * batch)
        df_part = spark.createDataFrame(df.iloc[low: up, :])

        if verbose:
            print(f'[{low}, {up}]')

        if i == 0 and reset:
            mode = 'overwrite'
        else:
            mode = 'append'

        spark_frame_to_hive(
            df_part, db_name=db_name, table_name=table_name, spark=spark, repartition=4,
            mode=mode, verbose=verbose)
        df_part.unpersist()







import numpy as np
import pandas as pd


def ratio(df: pd.DataFrame, dividend: str, divisor: str, new_col: str):
    df[new_col] = df[dividend] / df[divisor]


def cross_nan(item):
    if np.isnan(item[cols[0]]) and np.isnan(item[cols[1]]):
        return "A"
    elif np.isnan(item[cols[0]]) and (not np.isnan(item[cols[1]])):
        return "B"
    elif (not np.isnan(item[cols[0]])) and np.isnan(item[cols[1]]):
        return "C"
    else:
        return "D"


def cross_median(item, ):
    """
    median_a 月份数
    median_b 融资次数
    :param item:
    :return:
    """
    if (item[cols[0]] <= median_a) and (item[cols[1]] <= median_b):
        return 2
    elif (item[cols[0]] <= median_a) and (item[cols[1]] > median_b):
        return 3
    elif (item[cols[0]] > median_a) and (item[cols[1]] <= median_b):
        return 5
    elif (item[cols[0]] > median_a) and (item[cols[1]] > median_b):
        return 4
    else:
        return 1


def joint_feature(df, corss_type, med_a=None, med_b=None, dropna=False):
    global cols, median_a, median_b

    if dropna == True:
        df.dropna(axis=0, inplace=True)

    cols = df.columns.tolist()
    target_name = cols.pop(-1)
    if med_a is not None:
        median_a = med_a
        median_b = med_b
    else:
        median_a = df[cols[0]].median()
        median_b = df[cols[1]].median()
    if corss_type == 'nan':
        df['corr'] = df.apply(cross_nan, axis=1)
    elif corss_type == 'median':
        df['corr'] = df.apply(cross_median, axis=1)
    return df


class FeatureTools:

    def __init__(self, df, conf: [dict, str], spark):

        if isinstance(conf, dict):
            self.conf = conf
        elif isinstance(conf, str):
            self.conf = self._load_config(conf)

        self.min_value = 1e-6
        self.spark = spark
        self.df = df.orderBy('uuid', 'month')
        self.df_res = self.df.select('uuid').dropDuplicates()
        self.datas = []
        self.datas_aggs = []
        self.__gen_months()
        self.__gen_datas()
        print('Pre Code Done!')

    def __gen_months(self, ):
        start_month = pd.to_datetime(self.conf.get('observe_time') + '01') - pd.DateOffset(
            months=int(self.conf.get('months_cnt')) - 1)
        end_month = pd.to_datetime(self.conf.get('observe_time') + '01') + pd.DateOffset(months=1)
        months = pd.date_range(start_month, end_month, freq='M', closed='left').astype(str).tolist()
        self.conf['months'] = [month.replace('-', '')[: 6] for month in months]

    def __gen_datas(self):

        def month_to_observe_time(months):
            def f(month):
                return 11 - months.index(month)

            return F.udf(f, returnType=IntegerType())

        for filter_rule in self.conf.get('filter_rules'):
            act_type, values = list(filter_rule.items())[0]

            df_agg_rule = self.df.where(self.df[act_type].isin(values))

            self.datas.append(df_agg_rule)

            self.datas_aggs.append(
                df_agg_rule.groupBy(self.conf.get('group_cols')).agg(
                    {}.fromkeys(self.conf.get('basic_vars'), self.conf.get('agg_type'))
                )
            )

        for i, data_agg in enumerate(self.datas_aggs):
            # 修正字段名称
            for col in data_agg.columns:
                if '(' in col:
                    data_agg = data_agg.withColumnRenamed(col, col.split('(')[1].split(')')[0])

            data_agg = data_agg.withColumn(
                'time_mean',
                F.abs(data_agg[self.conf.get('basic_vars')][0] / data_agg[self.conf.get('basic_vars')][1]))

            # 差分
            window = Window.orderBy('month').partitionBy('uuid', 'direction')
            data_agg = data_agg.withColumn('amt_lag_1', F.lag(data_agg.amt_raw).over(window))
            data_agg = data_agg.withColumn('cnt_lag_1', F.lag(data_agg.cnt_raw).over(window))
            data_agg = data_agg.withColumn('amt_lag', data_agg.amt_raw - data_agg.amt_lag_1)
            data_agg = data_agg.withColumn('cnt_lag', data_agg.cnt_raw - data_agg.cnt_lag_1)
            # 距今月数
            data_agg = data_agg.withColumn("month_to_observe",
                                           month_to_observe_time(self.conf['months'])(data_agg.month))
            self.datas_aggs[i] = data_agg

    def _get_month_filter(self, time_interval: int):
        return self.conf['months'][-time_interval:]

    def cal_index(self, data, rule_prefix='R1'):

        months = self.conf['months']

        def continuous_month(months_ls):
            if len(months_ls) == 0:
                return 0
            elif len(months_ls) < 2:
                return 1
            else:
                months_ls = [month[4:] for month in sorted(months_ls)]
                diff = []
                for index in range(len(months_ls) - 1):
                    diff.append(int(months_ls[index]) - int(months_ls[index + 1]))
                diff = [1 if diff_value == -1 or diff_value == 11 else 0 for diff_value in diff]
                diff = ''.join([str(diff_value) for diff_value in diff])
                counts = [item.count('1') for item in diff.split('0')]
                return max(counts) + 1

        @F.udf(returnType=IntegerType())
        def non_pay_month(month_ls):
            max_month = max(month_ls)
            return 11 - months.index(max_month)

        @F.udf(returnType=IntegerType())
        def continuous_pay_month(months_ls):
            return continuous_month(months_ls)

        def continuous_no_pay_month(month_filter):
            def fun(months_ls):
                month_ls = list(set(month_filter) - set(months_ls))
                return continuous_month(month_ls)

            return F.udf(fun, returnType=IntegerType())

        def value_compare(values):
            if len(values) == 0:
                return [0]
            else:
                values = [float(value) for value in values]
                mean = sum(values) / len(values)
                values = [0 if (value - mean) < 0 else 1 for value in values]
                return values

        @F.udf(returnType=IntegerType())
        def more_than_cnt(values):
            return value_compare(values).count(1)

        @F.udf(returnType=IntegerType())
        def less_than_cnt(values):
            return value_compare(values).count(0)

        def trend_analysis(values):
            values = [float(value) for value in values]
            pres = values[: -1]
            tails = values[1:]
            return [tail - pre for pre, tail in zip(pres, tails)]

        @F.udf(returnType=IntegerType())
        def increase(values):
            if len(values) <= 1:
                return -1
            else:
                masks = trend_analysis(values)
                masks = [mask > 0 for mask in masks]
                return 1 if all(masks) else 0

        @F.udf(returnType=IntegerType())
        def decrease(values):
            if len(values) <= 1:
                return -1
            else:
                masks = trend_analysis(values)
                masks = [mask < 0 for mask in masks]
                return 1 if all(masks) else 0

        df_res = data.select('uuid').dropDuplicates()
        for time_interval, direction in product(self.conf.get('time_range_list'), self.conf.get('direction_type')):
            month_filter = self._get_month_filter(time_interval)
            cond = data.month.isin(month_filter) & (data[self.conf.get('direction')] == direction)
            part_data = data.where(cond)

            # month_filter_spark = self.spark.sparkContext.broadcast(month_filter)
            part_res = part_data.groupBy('uuid').agg(
                # 金额
                F.sum(part_data.amt_raw).alias(f'{rule_prefix}_{time_interval}M_{direction}_amt_sum'),
                F.max(part_data.amt_raw).alias(f'{rule_prefix}_{time_interval}M_{direction}_amt_max'),
                F.min(part_data.amt_raw).alias(f'{rule_prefix}_{time_interval}M_{direction}_amt_min'),
                F.avg(part_data.amt_raw).alias(f'{rule_prefix}_{time_interval}M_{direction}_amt_avg'),
                F.stddev(part_data.amt_raw).alias(f'{rule_prefix}_{time_interval}M_{direction}_amt_std'),

                # 次数
                F.sum(part_data.cnt_raw).alias(f'{rule_prefix}_{time_interval}M_{direction}_cnt_sum'),
                F.max(part_data.cnt_raw).alias(f'{rule_prefix}_{time_interval}M_{direction}_cnt_max'),
                F.min(part_data.cnt_raw).alias(f'{rule_prefix}_{time_interval}M_{direction}_cnt_min'),
                F.avg(part_data.cnt_raw).alias(f'{rule_prefix}_{time_interval}M_{direction}_cnt_avg'),
                F.stddev(part_data.cnt_raw).alias(f'{rule_prefix}_{time_interval}M_{direction}_cnt_std'),

                # 月份类
                # 总交易月份数
                F.countDistinct(part_data.month).alias(f'{rule_prefix}_{time_interval}M_{direction}_months_pay_cnt'),
                # 距今交易最近月份
                F.max(part_data.month).alias(f'{rule_prefix}_{time_interval}M_{direction}_months_rct'),
                # 距今未交易日期数，距今无交易月份数
                F.min(part_data.month_to_observe).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_months_rct_no_pay'),
                # 交易日期比例
                (F.countDistinct(part_data.month) / len(month_filter)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_months_pay_rate'),
                # 未交易日期数
                (len(month_filter) - F.countDistinct(part_data.month)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_months_nopay_cnt'),
                # 最大连续交易月数
                continuous_pay_month(F.collect_list(part_data.month)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_months_max_ctn_cnt'),
                # 最大连续不交易数
                continuous_no_pay_month(month_filter)(F.collect_list(part_data.month)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_months_max_nopay_ctn_cnt'),

                # 笔均金额
                (F.sum(part_data.amt_raw) / F.sum(part_data.cnt_raw)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_amt_atm_avg'),
                F.max(part_data.time_mean).alias(f'{rule_prefix}_{time_interval}M_{direction}_amt_atm_avg_max'),
                F.min(part_data.time_mean).alias(f'{rule_prefix}_{time_interval}M_{direction}_amt_atm_avg_min'),
                F.avg(part_data.time_mean).alias(f'{rule_prefix}_{time_interval}M_{direction}_amt_atm_avg_avg'),
                F.stddev(part_data.time_mean).alias(f'{rule_prefix}_{time_interval}M_{direction}_amt_atm_avg_std'),

                # 月交易金额小于月均金额月数
                less_than_cnt(F.collect_list(part_data.amt_raw)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_amt_less_than_avg'),
                # 月交易金额大于等于月均金额月数
                more_than_cnt(F.collect_list(part_data.amt_raw)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_amt_more_than_avg'),
                # 月交易次数小于月均次数
                less_than_cnt(F.collect_list(part_data.cnt_raw)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_cnt_less_than_avg'),
                # 月交易次数大于等于月均次数
                more_than_cnt(F.collect_list(part_data.cnt_raw)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_cnt_more_than_avg'),

                # 交易金额严格递增
                increase(F.collect_list(part_data.amt_raw)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_amt_abs_increase'),
                # 交易金额严格递减
                decrease(F.collect_list(part_data.amt_raw)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_amt_abs_decrease'),
                # 交易次数严格递增
                increase(F.collect_list(part_data.cnt_raw)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_cnt_abs_increase'),
                # 交易次数严格递减
                decrease(F.collect_list(part_data.cnt_raw)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_cnt_abs_decrease'),

                # 一阶差分，交易金额严格递增
                increase(F.collect_list(part_data.amt_lag)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_amt_lag_increase'),
                # 一阶差分，交易金额严格递减
                decrease(F.collect_list(part_data.amt_lag)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_amt_lag_decrease'),
                # 一阶差分，交易次数严格递增
                increase(F.collect_list(part_data.cnt_lag)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_cnt_lag_increase'),
                # 一阶差分，交易次数严格递减
                decrease(F.collect_list(part_data.cnt_lag)).alias(
                    f'{rule_prefix}_{time_interval}M_{direction}_cnt_lag_decrease'),
            )

            df_res = df_res.join(part_res, how='left', on='uuid')

        # 近X月次数或金额比近12月次数或金额
        for time_interval, direction in product(self.conf.get('time_range_list')[:-1], self.conf.get('direction_type')):
            if time_interval * 2 <= self.conf.get('time_range_list')[-1]:
                # 2TR
                df_res = df_res.withColumn(f'{rule_prefix}_{time_interval}M_{direction}_amt_sum_div2tr',
                                           df_res[f'{rule_prefix}_{time_interval}M_{direction}_amt_sum'] / (df_res[
                                                                                                                f'{rule_prefix}_{time_interval * 2}M_{direction}_amt_sum'] - self.min_value))
                df_res = df_res.withColumn(f'{rule_prefix}_{time_interval}M_{direction}_cnt_sum_div2tr',
                                           df_res[f'{rule_prefix}_{time_interval}M_{direction}_cnt_sum'] / (df_res[
                                                                                                                f'{rule_prefix}_{time_interval * 2}M_{direction}_cnt_sum'] - self.min_value))
                # 环比
                df_res = df_res.withColumn(f'{rule_prefix}_{time_interval}M_{direction}_amt_sum_chain_rate',
                                           df_res[f'{rule_prefix}_{time_interval}M_{direction}_amt_sum'] / (df_res[
                                                                                                                f'{rule_prefix}_{time_interval * 2}M_{direction}_amt_sum'] -
                                                                                                            df_res[
                                                                                                                f'{rule_prefix}_{time_interval}M_{direction}_amt_sum'] - self.min_value))
                df_res = df_res.withColumn(f'{rule_prefix}_{time_interval}M_{direction}_cnt_sum_chain_rate',
                                           df_res[f'{rule_prefix}_{time_interval}M_{direction}_cnt_sum'] / (df_res[
                                                                                                                f'{rule_prefix}_{time_interval * 2}M_{direction}_cnt_sum'] -
                                                                                                            df_res[
                                                                                                                f'{rule_prefix}_{time_interval}M_{direction}_cnt_sum'] - self.min_value))

            df_res = df_res.withColumn(f'{rule_prefix}_{time_interval}M_{direction}_amt_sum_div1y',
                                       df_res[f'{rule_prefix}_{time_interval}M_{direction}_amt_sum'] / (
                                                   df_res[f'{rule_prefix}_12M_{direction}_amt_sum'] - self.min_value))
            df_res = df_res.withColumn(f'{rule_prefix}_{time_interval}M_{direction}_cnt_sum_div1y',
                                       df_res[f'{rule_prefix}_{time_interval}M_{direction}_cnt_sum'] / (
                                                   df_res[f'{rule_prefix}_12M_{direction}_cnt_sum'] - self.min_value))

        return df_res

    def cal_indexs(self):
        for filter_rule in range(len(self.conf.get('filter_rules'))):
            df = self.datas_aggs[filter_rule]
            res = self.cal_index(data=df, rule_prefix=f'R{filter_rule}')
            self.df_res = self.df_res.join(res, how='left', on='uuid')
            print(f'Rule {filter_rule + 1} Done! With {len(res.columns)} Columns!')
        print(f'ALL {len(self.df_res.columns)} Columns!')

    def _save_config(self):
        Pk.save(self.conf.get('config_file_name.pk'))

    def _load_config(self, conf):
        return Pk.load(conf)

    def _set_filter_rules(self, filter_rules):
        self.conf['filter_rules'] = filter_rules

    def _set_agg_type(self, agg_type):
        self.conf['agg_type'] = agg_type




import numpy as np
import pandas as pd
from pandas import DataFrame
from typing import Any, Dict, Tuple, Callable, Sequence, Optional
from ..utils.logger import LoggerMixin


__all__ = ['AHP']


class AHP(LoggerMixin):
    """"""
    df_weight: DataFrame

    def __init__(
            self,
            metrix: DataFrame,
            value: Optional[str] = 'value',
            control: Optional[str] = 'control',
            experimental: Optional[str] = 'experimental',
            logger: Optional[Callable] = None,
            verbose: Optional[bool] = True,
            *args: Any,
            **kwargs: Any,
    ):
        """"""
        self.metrix = metrix
        self.value = value
        self.control = control
        self.experimental = experimental
        self.metrix_weight = dict()
        self.metrix_data = dict()
        self.group_ci = dict()
        self.group_cr = dict()
        self.ri = {
            1: 1e-6, 2: 1e-6, 3: 0.58, 4: 0.9, 5: 1.12, 6: 1.24, 7: 1.32, 8: 1.41, 9: 1.45, 10: 1.49,
            11: 1.51, 12: 1.54, 13: 1.56, 14: 1.58, 15: 1.59, 16: 1.60,
        }
        self.logger = logger
        self.verbose = verbose
        self.args = args
        self.kwargs = kwargs
        self._validate_data()

    def _validate_data(self):
        """"""
        for column in [self.value, self.control, self.experimental]:
            if column not in self.metrix.columns:
                raise KeyError(f"Column {column} not found in metrix.")

        for group_name, df_group in self.metrix.groupby("group"):
            group_shape = df_group.shape
            group_index_len = df_group[self.control].nunique()
            if group_shape[0] != group_index_len ** 2:
                raise ValueError(f"Metrix <{group_name}> shape {group_shape[0]} is not equal to {group_index_len ** 2}")

    def geometric_mean(self, numbers: Sequence) -> float:
        """"""
        return np.power(np.prod(numbers), 1 / len(numbers))

    def _ci_cr(self, metrix: DataFrame, metrix_weight: DataFrame, metrix_name: str) -> Tuple[float, float]:
        """
        计算一致性指标CI
        CI = (λ_avg - n) / (n - 1)
        CR = CI / RI
        """
        lambda_avg = (metrix.dot(metrix_weight.weight) / metrix_weight.weight).mean()
        _len = len(metrix_weight)
        ci = (lambda_avg - _len) / (_len - 1)
        if _len <= 2:
            cr = 0
        else:
            cr = ci / self.ri.get(_len)
        self.group_ci.update({metrix_name: ci})
        self.group_cr.update({metrix_name: cr})
        if cr > 0.1:
            self._logger(msg=f"[{__class__.__name__}] Metrix: <{metrix_name}> CR = {cr:.2f} CR is too large, please check your metrix.", color="red")
        else:
            self._logger(msg=f"[{__class__.__name__}] Metrix: <{metrix_name}> CR = {cr:.2f}", color="green")
        return ci, cr

    def solve_metrix(self):
        for group_name, df_group in self.metrix.groupby("group"):
            metrix = df_group.pivot_table(values="value", columns=['control'], index=["experimental"])
            metrix_weight = pd.DataFrame(
                metrix.apply(self.geometric_mean, axis=1),
                columns=["geo_mean"]
            )
            metrix_weight["weight"] = metrix_weight.geo_mean / metrix_weight.geo_mean.sum()
            ci, cr = self._ci_cr(metrix, metrix_weight, metrix_name=str(group_name))
            self.metrix_data.update({group_name: metrix})
            self.metrix_weight.update({group_name: metrix_weight})

    def weight(
            self,
    ) -> DataFrame:
        """"""
        if len(self.metrix_weight) == 0:
            raise ValueError("Please solve metrix first.")
        df_weight = pd.concat([weight.reset_index(drop=False) for key, weight in self.metrix_weight.items()])
        df_weight.rename(columns={"experimental": "dst", "weight": "group_weight"}, inplace=True)
        self.df_weight = df_weight.reset_index(drop=True)
        return self.df_weight



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns


sns.set(style="white")


def diagonal_corr(df):
    # Compute the correlation matrix
    corr = df.corr()

    # Generate a mask for the upper triangle
    mask = np.triu(np.ones_like(corr, dtype=np.bool))

    # Set up the matplotlib figure
    f, ax = plt.subplots(figsize=(11, 9))

    # Generate a custom diverging colormap
    cmap = sns.diverging_palette(220, 10, as_cmap=True)

    # Draw the heatmap with the mask and correct aspect ratio
    sns.heatmap(corr, mask=mask, cmap=cmap, vmax=.3, center=0,
                square=True, linewidths=.5, cbar_kws={"shrink": .5})


def structure_corr(df, col_cate, target_name='target'):
    '''
    col_cate = [f"{item.split('_')[0]}_{item.split('_')[1]}" for item in col]
    :param df:
    :param col_cate:
    :return:
    '''
    col_cate = sorted(col_cate)
    col = sorted(df.columns.tolist())
    df = df[col]
    col_cate_unique = np.unique(col_cate)
    col_cate_values = pd.Series(col_cate).replace(col_cate_unique, range(len(col_cate_unique))).values

    # Create a categorical palette to identify the networks
    pal = sns.husl_palette(len(col_cate), s=.45)
    lut = dict(zip(np.unique(col_cate_values), pal))
    colors = [lut[item] for item in col_cate_values]

    sns.clustermap(
        df.corr(), center=0, cmap="vlag",
        row_colors=colors, col_colors=colors,
        linewidths=.75, figsize=(18, 18)
    )






import os
import sys
import json
import time
import pickle
import logging
import random
import pandas as pd
from typing import Any, List, Optional, Sequence


__all__ = [
    "batches",
    "dispatch_tqdm",
    'WarnInfo',
    'getNowDate',
    'convertColToList',
    'convertDictToDataFrame',
    'getFolders',
    'saveCsv',
    'sortColumns',
    'mkdir',
    'Pk',
    'Json',
]


def batches(
        lst: list,
        batch_size: int,
) -> List:
    """
    desc: 生成批次数据
    :param lst: 原始List
    :param batch_size: 批次大小
    :return:
    """
    for i in range(0, len(lst), batch_size):
        yield lst[i:i+batch_size]


def dispatch_tqdm(items: Sequence, desc: Optional[str] = None) -> Any:
    """"""
    if 'ipython' in sys.argv[0].lower() or "ipykernel" in sys.argv[0].lower():
        from tqdm import tqdm_notebook
        pbar = tqdm_notebook(items, desc=desc)
    else:
        from tqdm import tqdm
        pbar = tqdm(items, desc=desc)
    return pbar


def WarnInfo(info, disp=True):
    """
    描述：提示一些信息。
    """
    if disp:
        logging.warning(info)


def getNowDate():
    '''
    描述：返回当前日期
    format[YYYYMMDD]: 20200202

    示例：getNowDate()
    '''
    return time.strftime("%Y%m%d", time.localtime())

def convertColToList(df, col):
    """
    描述：convert pyspark DataFrame to list.

    参数：
    :param df: 需要转换的 df[PySpark DataFrame]
    :param col: 需要转换的字段名称 column name

    示例：

    """
    return df.select(col).rdd.map(lambda x: x[0]).collect()


def getFolders(path):
    """
    描述：返回路径下全部的文件夹
    """
    files = os.listdir(path)
    Folders = []
    for folder in files:
        if '.' not in folder and '_' not in folder:
            Folders.append(folder)
    return Folders


def saveCsv(df, csvName):
    """

    """
    if not os.path.exists('res'):
        os.mkdir('res')
    df.to_csv(os.path.join('res', csvName + '.csv'), index=False)

def convertDictToDataFrame(dataType):

    newDataType = {}
    lengthList = []

    for item in dataType.items():
        lengthList.append(len(item[1]))
    maxLength = max(lengthList)
    for key in dataType.keys():
        newDataType[key] = dataType.get(key) + [None] * (maxLength - len(dataType.get(key)))
    return pd.DataFrame(newDataType)

def sortColumns(dfCount) -> list:
    """
    给定多列不同列长的 df.count() 进行列的排序，形成上三角排列。

    :param dfCount: Series df.count()
    :return: new columns name
    """

    _dfNum = pd.DataFrame(dfCount, columns=['num'])
    _newCols = [sorted(group.index.to_list()) for i, group in _dfNum.groupby('num', sort=True)][::-1]
    newCols = []
    for item in _newCols:
        newCols += item
    return newCols


def mkdir(filename):
    """
    描述：检查文件路径是否存在，不存在的情况下创建路径。

    参数：
    :param filename: 文件路径
    :return: None
    """
    path = os.path.dirname(filename)
    if not os.path.exists(path):
        os.mkdir(path)

class Pk:
    """
    描述：写入与读取 `.pickle` 文件
    """
    @staticmethod
    def save(filename, data):
        """
        描述：保存数据至 .pickle 文件

        参数：
        :param filename: 保存文件路径
        :param data: 需要保存的数据
        :return: None
        """
        mkdir(filename)
        with open(filename, 'wb+') as f:
            pickle.dump(data, f)

    @staticmethod
    def load(filename):
        """
        描述：载入 .pickle 数据

        参数：
        :param filename: 文件路径
        :return: data .pickle 文件中的数据
        """
        with open(filename, 'rb') as f:
            data = pickle.load(f, encoding='gbk')
        return data

class Json:
    """
    描述：写入与读取 `.json` 文件
    """

    @staticmethod
    def save(filename, data):
        """
        描述：保存数据至 .json 文件

        参数：
        :param filename: 保存文件路径
        :param path: 需要保存的数据
        :return: None
        """
        mkdir(filename)
        data = json.dumps(data, ensure_ascii=False, indent=4)
        with open(filename, 'w+') as f:
            f.write(data)

    @staticmethod
    def load(filename):
        """
        描述：载入 .json 数据

        :param path: 文件路径
        :return: data .json 文件中的数据
        """
        with open(filename, "r", encoding="utf-8") as f:
            data = json.load(f)
        return data


def decompose_number(
        number:float,
        values:int
) -> List[float]:
    """
    desc: 分解一个数值，将一个数值分解为N个数值
    :param target:
    :param num_values:
    :return:
    >>> number = 100
    >>> values = 5
    >>> random_numbers = decompose_number(number, values)
    """
    decompose_values = []
    total = 0

    for _ in range(values - 1):
        value = random.uniform(0, number - total)
        decompose_values.append(value)
        total += value

    decompose_values.append(number - total)
    random.shuffle(decompose_values)
    return decompose_values
#
# def do_work(img_path, use_gpu=False):
#     """"""
#     result = ocr.ocr(img_path)
#     return result
#
# def run_with_mp_map(items, do_work, processes=None, chunksize=1):
#     """"""
#     with Pool(processes=processes) as pool:
#         results = pool.map(do_work, items, chunksize=chunksize)
#     return results




from scipy.stats import beta


__all__ = [
    "BetaDistribution"
]


class BetaDistribution:

    def __init__(self, Alpha: float = None, Beta: float = None):
        self.Alpha = Alpha
        self.Beta = Beta

    def grade_rate(self, target_grade: int):
        """
        计算 Beta 分布下，每个评级等级的比率
        :return:
        """
        target_grade_rate, CDF = [], []
        for i in range(target_grade):
            F_left = beta.cdf(x=(i / (target_grade)), a=self.Alpha, b=self.Beta, loc=0, scale=1)
            F_right = beta.cdf(x=((i + 1) / (target_grade)), a=self.Alpha, b=self.Beta, loc=0, scale=1)
            target_grade_rate.append(F_right - F_left)
            CDF.append(F_right)
        return target_grade_rate, CDF



import os
import json
import pandas as pd
from IPython.display import display, Markdown


class MarkDown:


    def __init__(self, fileName='', filePath='./'):
        self.MarkDownContent = []
        self.fileName = fileName
        self.filePath = filePath
        self.fileAddress = os.path.join(self.filePath, self.fileName)

    def __ifCenter(self, string):
        return '<center>' + string + '</center>'


    def addLine(self, lineContent, rowNum=False, ifCenter=False):
        if not rowNum:
            if ifCenter:
                lineContent = self.__ifCenter(lineContent)
            self.MarkDownContent.append(lineContent + ' \n')


    def addImage(self, imagePath, box=['', ''], ifCenter=True):
        lineContent = '<img src="{}" width="{}" height="{}">'.format(imagePath, box[0], box[1])
        self.addLine(lineContent, ifCenter=ifCenter)


    def addTable(self, df, ifCenter=True):
        self.addLine(df.to_html(), ifCenter=True)


    def addFomula(self, fomula):
        self.addLine('$$\n' + fomula + '\n$$')


    def showMDinNotebook(self, ):
        return display(Markdown(''.join(self.MarkDownContent)))


    def saveMD(self, fileName='newMD.md', filePath='./'):
        fileAddress = os.path.join(filePath, fileName)
        with open(fileAddress, "w") as file:
            file.write('\n'.join(self.MarkDownContent))
        print("\033[32mFile {} saved!\033[0m".format(fileAddress))


    def readMD(self, ):
        if os.path.exists(self.fileAddress):
            with open(self.fileAddress, 'r+') as MarkDownFile:
                self.MarkDownContent = MarkDownFile.readlines()
        else:
            print("\033[32mFile {} does not exist!\033[0m".format(self.fileAddress))



import numpy as np
import pandas as pd
from statsmodels.api import OLS
from tqdm import tqdm_notebook

from pyspark.ml.feature import VectorAssembler
from pyspark.ml.regression import LinearRegression
from statsmodels.stats.outliers_influence import variance_inflation_factor as vif


class VIF:
    """
    全部VIF方法
    1. statsmodels `vif()`            statsmodels VIF
    2. statsmodels `StatsTableVIF()`  statsmodels 计算全表VIF
    3. statsmodels `LowVIFFeatures()` 依据VIF设定阈值，选择最小VIF特征组合
    4. PySpark     `SparkVIF()`       PySpark 计算变量VIF
    5. PySpark     `SparkTableVIF()`  PySpark 计算全表VIF
    """
    @staticmethod
    def vif(exog, exog_idx):
        """
        改进 statsmodels 中的 variance_inflation_factor
        :param exog:
        :param exog_idx:
        :return:
        """
        k_vars = exog.shape[1]
        x_i = exog[:, exog_idx]
        mask = np.arange(k_vars) != exog_idx
        x_noti = exog[:, mask]
        r_squared_i = OLS(x_i, x_noti, missing='drop').fit().rsquared
        return 1. / (1. - r_squared_i)

    @staticmethod
    def StatsTableVIF(df):
        """
        描述：statsmodels `StatsTableVIF()` 计算全表 VIF

        参数：
        :param df[pandas.DataFrame]: 全量需要计算`VIF`的特征`DataFrame`

        示例：
        >>> pd.DataFrame(VIF.StatsTableVIF(df))
        """
        feature_vif = []
        for i, column_name in enumerate(df.columns):
            feature_vif.append({
                'feature': column_name,
                'vif_value': vif(df.values, i)
            })
        return feature_vif

    @staticmethod
    def LowVIFFeatures(df, thresh=5):
        """
        描述：选择最低VIF特征组合。

        参数：
        :param df[pandas.DataFrame]: 特征
        :param thresh[int]: VIF阈值
        :return [pandas.DataFrame]: 指标名称列表，筛选明细

        示例：
        >>> VIF.LowVIFFeatures(df, thresh=5)
        """
        feature_vif = VIF.StatsTableVIF(df)
        df_vif = pd.DataFrame(feature_vif)
        max_vif = df_vif.vif_value.max()

        while max_vif >= thresh and len(df.columns) > 2:
            vif_sums = []
            vif_maxs = []
            vif_ress = []

            for col in df.columns:
                new_vif = pd.DataFrame(VIF.StatsTableVIF(df.drop(col, axis=1)))
                vif_sums.append(new_vif.vif_value.sum())
                vif_maxs.append(new_vif.vif_value.max())
                vif_ress.append(new_vif)

            min_index = vif_sums.index(min(vif_sums))
            max_vif = vif_maxs[min_index]
            vif_res = vif_ress[min_index]
            drop_col = df.columns[min_index]
            df.drop(drop_col, axis=1, inplace=True)
            df_vif = df_vif.merge(vif_res, on='feature', how='left')
            features = df_vif.iloc[:, -1].dropna().shape[0]
            print(f'features: {features}')

        columns_name = ['feature'] + [f'step_{i}' for i in range(1, len(df_vif.columns))]
        df_vif.columns = columns_name
        return df_vif

    @staticmethod
    def SparkVIF(df, var_name, other_vars=None):
        """
        描述：计算方差膨胀因子

        参数：
        :param df[PySpark DataFrame]: 数据表 PySpark DataFrame
        :param var_name[str]: 需要计算 VIF 的字段名称
        :param other_vars[list]: 需要计算 VIF 的参照变量，默认为 `df` 中除了 `var_name` 的全部变量
        :return vif[float]: 变量 `var_name` 的方差膨胀因子

        示例：
        >>> VIF.SparkVIF(df, 'col_0')
        """
        if other_vars is None:
            other_vars = df.drop(var_name).columns

        vector_translate = VectorAssembler(inputCols=other_vars, outputCol='features')
        df_feature = vector_translate.transform(df)

        linear = LinearRegression(
            labelCol=var_name,
            fitIntercept=False,
        ).fit(df_feature)

        r_squared = linear.summary.r2
        return 1. / (1. - r_squared)

    @staticmethod
    def SparkTableVIF(df):
        """
        描述：计算整个Spark DataFrame表的方差膨胀因子。

        :param df: Spark DataFrame
        :return: Pandas DataFrame
        """
        columns_name = df.columns
        vifs = []
        for column_name in tqdm_notebook(columns_name):
            vifs.append(VIF.SparkVIF(df, var_name=column_name))
        return pd.DataFrame({
            'var_name': columns_name,
            'VIF': vifs,
        })





from typing import Callable
from pydantic import Field, BaseModel, ConfigDict
try:
    from pyspark.sql import functions as F
except ImportError:
    print("pyspark not installed, please install it first. < pip install pyspark >")


__all__ = [
    "SparkBaseOperator",
    "spark_base_operator",
]


class SparkBaseOperator(BaseModel):
    """base operator class for spark"""
    model_config = ConfigDict(arbitrary_types_allowed=True)
    sum: Callable = Field(default=F.sum)
    avg: Callable = Field(default=F.avg)
    mean: Callable = Field(default=F.mean)
    count: Callable = Field(default=F.count)
    count_distinct: Callable = Field(default=F.countDistinct)
    sum_distinct: Callable = Field(default=F.sumDistinct)
    min: Callable = Field(default=F.min)
    max: Callable = Field(default=F.max)
    first: Callable = Field(default=F.first)
    last: Callable = Field(default=F.last)
    skewness: Callable = Field(default=F.skewness)
    kurtosis: Callable = Field(default=F.kurtosis)
    collect_set: Callable = Field(default=F.collect_set)
    collect_list: Callable = Field(default=F.collect_list)
    approx_count_distinct: Callable = Field(default=F.approx_count_distinct)
    stddev: Callable = Field(default=F.stddev)
    stddev_samp: Callable = Field(default=F.stddev_samp)
    stddev_pop: Callable = Field(default=F.stddev_pop)
    variance: Callable = Field(default=F.variance)
    var_pop: Callable = Field(default=F.var_pop)
    var_samp: Callable = Field(default=F.var_samp)


spark_base_operator = SparkBaseOperator()



try:
    from colorama import Fore, Style
except ModuleNotFoundError:
    raise ModuleNotFoundError("pip install colorama")
from typing import Dict, Optional, Callable, Literal


__all__ = ["LoggerMixin"]


print_color_mapping = {
    "BLACK".lower(): Fore.BLACK,
    "RED".lower(): Fore.RED,
    "GREEN".lower(): Fore.GREEN,
    "YELLOW".lower(): Fore.YELLOW,
    "BLUE".lower(): Fore.BLUE,
    "CYAN".lower(): Fore.CYAN,
    "MAGENTA".lower(): Fore.MAGENTA,
    "WHITE".lower(): Fore.WHITE,
    "RESET".lower(): Fore.RESET,
}


TypePrintColor = Literal['black', 'red', 'green', 'yellow', 'blue', 'cyan', 'magenta', 'white', 'reset']


class LoggerMixin:
    """"""
    logger: Optional[Callable]
    verbose: Optional[bool]

    def _logger(
            self,
            msg: str,
            color: Optional[TypePrintColor] = None
    ) -> None:
        """"""
        if self.logger: self.logger.info(msg)
        if self.verbose:
            if color:
                print(print_color_mapping.get(color) + msg + Style.RESET_ALL, flush=True)
            else:
                print(msg, flush=True)

    def _logger_base(self, msg: str, color: Optional[TypePrintColor] = None):
        """"""
        if self.logger: self.logger.info(msg)
        if self.verbose: print(print_color_mapping.get(color) + msg + Style.RESET_ALL, flush=True)

    def _logger_agent_start(
            self,
            msg: Optional[str] = None,
            name: Optional[str] = None,
            color: Optional[TypePrintColor] = "green"
    ):
        """"""
        if msg is None: msg = f"[{name}] Start ...\n"
        self._logger_base(msg=msg, color=color)

    def _logger_agent_end(
            self,
            msg: Optional[str] = None,
            name: Optional[str] = None,
            color: Optional[TypePrintColor] = "green"
    ):
        """"""
        if msg is None: msg = f"[{name}] End ...\n"
        self._logger_base(msg=msg, color=color)

    def _logger_agent_question(
            self,
            msg: Optional[str] = None,
            name: Optional[str] = None,
            content: Optional[str] = None,
            color: Optional[TypePrintColor] = "green"
    ):
        """"""
        if msg is None: msg = f"[{name}] User Question: {content}\n"
        self._logger_base(msg=msg, color=color)

    def _logger_messages_start(
            self,
            msg: Optional[str] = None,
            name: Optional[str] = None,
            color: Optional[TypePrintColor] = "red"
    ):
        """"""
        if msg is None: msg = f"{20 * '='} [{name}] Messages Start {20 * '='}\n"
        self._logger_base(msg=msg, color=color)

    def _logger_messages_end(
            self,
            msg: Optional[str] = None,
            name: Optional[str] = None,
            color: Optional[TypePrintColor] = "red"
    ):
        """"""
        if msg is None: msg = f"{20 * '='} [{name}] Messages End    {20 * '='}\n"
        self._logger_base(msg=msg, color=color)

    def _logger_messages(
            self,
            msg: Optional[str] = None,
            role: Optional[str] = None,
            content: Optional[str] = None,
            color: Optional[TypePrintColor] = "blue"
    ):
        """"""
        if msg is None: msg = f"{role}: [{content}]\n"
        self._logger_base(msg=msg, color=color)

    def _logger_agent_script(
            self,
            msg: Optional[str] = None,
            name: Optional[str] = None,
            script: Optional[str] = None,
            color: Optional[TypePrintColor] = "magenta"
    ):
        """"""
        if msg is None: msg = f"[{name}] Script: ```\n{script}\n```"
        self._logger_base(msg=msg, color=color)

    def _logger_agent_search(
            self,
            msg: Optional[str] = None,
            name: Optional[str] = None,
            content: Optional[str] = None,
            color: Optional[TypePrintColor] = "magenta"
    ):
        """"""
        if msg is None: msg = f"[{name}] Script: ```\n{content}\n```"
        self._logger_base(msg=msg, color=color)

    def _logger_agent_warning(self, msg: str, color: Optional[TypePrintColor] = "red"):
        """"""
        self._logger_base(msg=msg, color=color)

    def _logger_agent_final_answer(
            self,
            msg: Optional[str] = None,
            name: Optional[str] = None,
            content: Optional[str] = None,
            color: Optional[TypePrintColor] = "yellow"
    ):
        """"""
        if msg is None: msg = f"[{name}] Final Answer: {content}\n"
        self._logger_base(msg=msg, color=color)

    def _logger_dict(self, msg: Dict, color="green"):
        """"""
        for key, val in msg.items():
            if self.logger: self.logger.info(f"{key}: {val}")
            if self.verbose: print(print_color_mapping.get(color) + f"{key}: {val}" + Style.RESET_ALL, flush=True)

    def _logger_color(self, msg: str) -> None:
        """"""
        last_line = msg.split('\n')[-1]

        if self.logger: self.logger.info(msg)
        if self.verbose:
            if last_line.startswith("Answer"):
                print(Fore.YELLOW + msg + Style.RESET_ALL)
            elif last_line.startswith("Action"):
                print(Fore.GREEN + msg + Style.RESET_ALL)
            elif last_line.startswith("No Action"):
                print(Fore.RED + msg + Style.RESET_ALL)
            elif last_line.startswith("Observation"):
                print(Fore.BLUE + msg + Style.RESET_ALL)
            elif last_line.startswith("Running"):
                print(Fore.CYAN + msg + Style.RESET_ALL)
            else:
                print(msg)
            print()



import numpy as np
import os

from sklearn.preprocessing import label_binarize
from sklearn.utils import column_or_1d, check_consistent_length
from numpy import log
from pandas import cut, DataFrame
from pandas.api.types import is_numeric_dtype
from kivi.utils import saveCsv
from kivi.evaluate import utils


class Calibration:
    def __init__(self, imageName=None):
        self.imageName = imageName

    def CalibrationCurve(self, y_true, y_prob, normalize=False, n_bins=5,
                         strategy='uniform'):
        """计算校正曲线(Calibration curve)的真实和预测概率。
            该方法假定输入来自二进制分类器。
            校准曲线也可以称为可靠性图(reliability diagrams)。

        参数
        ----------
        y_true : array, shape (n_samples,)
            True targets.
        y_prob : array, shape (n_samples,)
            Probabilities of the positive class.
        normalize : bool, optional, default=False
            是否需要将y_prob归一化为bin [0，1]，不是适当的概率。
            如果为True，则y_prob中的最小值映射到0，最大的映射到1。
        n_bins : int
            分箱数。更大的数量需要更多的数据。没有数据点的bin（即y_prob中没有相应的值）将不会返回，
            因此返回值中的n_bins可能少于n_bins。
        strategy : {'uniform', 'quantile'}, (default='uniform')
            uniform: 等距
            quantile: 等频
        返回值
        -------
        prob_true : array, shape (n_bins,) or smaller
            每个分箱中的真实概率 (fraction of positives).
        prob_pred : array, shape (n_bins,) or smaller
            每个分箱中的平均预测概率。
        """

        y_true = column_or_1d(y_true)
        y_prob = column_or_1d(y_prob)
        check_consistent_length(y_true, y_prob)

        if normalize:  # Normalize predicted values into interval [0, 1]
            y_prob = (y_prob - y_prob.min()) / (y_prob.max() - y_prob.min())
        elif y_prob.min() < 0 or y_prob.max() > 1:
            raise ValueError(
                "y_prob has values outside [0, 1] and normalize is "
                "set to False.")

        labels = np.unique(y_true)
        if len(labels) > 2:
            raise ValueError("Only binary classification is supported. "
                             "Provided labels %s." % labels)
        y_true = label_binarize(y_true, labels)[:, 0]

        if strategy == 'quantile':  # Determine bin edges by distribution of data
            quantiles = np.linspace(0, 1, n_bins + 1)
            bins = np.percentile(y_prob, quantiles * 100)
            bins[-1] = bins[-1] + 1e-8
        elif strategy == 'uniform':
            bins = np.linspace(0., 1. + 1e-8, n_bins + 1)
        else:
            raise ValueError("Invalid entry to 'strategy' input. Strategy "
                             "must be either 'quantile' or 'uniform'.")

        binids = np.digitize(y_prob, bins) - 1

        bin_sums = np.bincount(binids, weights=y_prob, minlength=len(bins))
        bin_true = np.bincount(binids, weights=y_true, minlength=len(bins))
        bin_total = np.bincount(binids, minlength=len(bins))

        nonzero = bin_total != 0
        prob_true = (bin_true[nonzero] / bin_total[nonzero])
        prob_pred = (bin_sums[nonzero] / bin_total[nonzero])

        return prob_true, prob_pred

    def Plot(self, y_true, predictLs, bins=10, normalize=False,
             strategy='uniform', **kwargs):
        """
        Des: 绘制概率校准曲线
        :param y_true: 真实标签
        :param predictLs: [(name, predict), (name, predict), ……]
        :param bins: 分箱数
        :param normalize : bool, optional, default=False
            是否需要将y_prob归一化为bin [0，1]，不是适当的概率。
            如果为True，则y_prob中的最小值映射到0，最大的映射到1。
        :param strategy : {'uniform', 'quantile'}, (default='uniform')
            uniform: 等距
            quantile: 等频
        :param kwargs: 图像参数['style', 'figsize', 'dpi', 'labelfontsize', title, 'x_label', 'y_label'……]
        :return: None
        """

        kwargs.setdefault('style', 'seaborn')
        kwargs.setdefault('figsize', (10, 10))
        kwargs.setdefault('dpi', 150)
        kwargs.setdefault('labelfontsize', 10)
        kwargs.setdefault('tickfontsize', 8)
        kwargs.setdefault('titlefontsize', 12)
        kwargs.setdefault('x_label', 'Mean predicted value')
        kwargs.setdefault('y_label', "Fraction of positives")
        kwargs.setdefault('title', 'Calibration plots  (reliability curve)')

        import matplotlib.pyplot as plt

        plt.figure(figsize=kwargs.get('figsize'), dpi=kwargs.get('dpi'))
        ax1 = plt.subplot2grid((3, 1), (0, 0), rowspan=2)
        ax2 = plt.subplot2grid((3, 1), (2, 0))

        ax1.plot([0, 1], [0, 1], "k:", label="Perfectly calibrated")
        for name, predict in predictLs:
            fraction_of_positives, mean_predicted_value = self.CalibrationCurve(
                y_true,
                predict,
                n_bins=bins,
                normalize=normalize,
                strategy=strategy,
            )

            ax1.plot(mean_predicted_value, fraction_of_positives, "s-",
                     label="%s" % (name,))

            ax2.hist(predict, range=(0, 1), bins=10, label=name,
                     histtype="step", lw=2)

        ax1.set_ylabel(kwargs.get('y_label'))
        ax1.set_ylim([-0.05, 1.05])
        ax1.legend(loc="lower right")
        ax1.set_title(kwargs.get('title'))

        ax2.set_xlabel(kwargs.get('x_label'))
        ax2.set_ylabel("Count")
        ax2.legend(loc="upper center", ncol=2)

        plt.tight_layout()

        if self.imageName:
            if not os.path.exists('img'):
                os.mkdir('img')
            plt.savefig(f'./img/{self.imageName}_calibration.png', )
            return None
        else:
            return plt.show()

class ODDS(Bins):

    def __init__(self, csvName=None):
        """
        Des: 计算PSI
        :param margin: cutoff point 左右填充预留边界
        :param csvName: 结果保存的csv文件名称（如果需要保存的话）
        """
        super(ODDS, self).__init__()
        self.csvName = csvName

    def _odds(self, res):
        """
        Des: 计算PSI的核心函数
        :param res: 明细表
        :return: 明细表以及PSI值 res, psi
        """

        res['Expected_percentage'] = res.Expected / res.Expected.sum()
        res['target_percentage'] = res.target / res.target.sum()
        res['A-E'] = res.target_percentage - res.Expected_percentage
        res['A/E'] = res.target_percentage / res.Expected_percentage
        res['ln(A/E)'] = log(res['A/E'])
        res['index'] = res['A/E'] * res['ln(A/E)']
        psi = res['index'].sum()
        return res, psi

    def odds(self, score, target, bins=10, binType='distance', cutPoint=None):
        """
        Des: 连续型变量的PSI指标
        :param score: 预期分布
        :param target: 实际分布
        :param bins: 分箱数目
        :param binType: 分箱类型 'distance'等距 or 'frequency'等频 or 'customize'自定义
        :param csvName: 保存CSV Name, 默认不保存
        :return: 明细表以及PSI值 res, PSI
        """
        score.dropna(inplace=True)
        target.dropna(inplace=True)

        scoreMax = score.max()
        scoreMin = score.min()

        if not is_numeric_dtype(score):
            raise Exception(TypeError, 'score must be numeric!')

        if not is_numeric_dtype(target):
            raise Exception(TypeError, 'target must be numeric!')

        if binType == 'distance':
            cutoffPoint = self._cutoffPoint(
                bins=bins, min=scoreMin, max=scoreMax)
        elif binType == 'frequency':
            cutoffPoint = self._cutoffPoint(
                bins=bins, percent=True, score=score)
        elif binType == 'customize' and cutPoint:
            cutoffPoint = cutPoint
        else:
            raise Exception(TypeError, '分箱错误!')

        total, bad = target.count(), target.sum()
        good = total - bad

        Bucket = DataFrame({
            'score': score,
            'target': target,
            'bucket': cut(score, cutoffPoint, )
        }).groupby('bucket', as_index=True)

        res = DataFrame({
            'min_bin': Bucket.score.min(),
            'max_bin': Bucket.score.max(),
            'total': Bucket.target.count(),
            'bad': Bucket.target.sum(),
        })

        res['good'] = res.total - res.bad
        res['odd_bad/good'] = res.bad / res.good
        res['bad_rate'] = res['bad'] / res['total']  # 每个箱体中好样本所占总样本数的比例
        res['Ln(Odds)'] = log(res['odd_bad/good'])
        return res
        self.res, self.psi = self._odds(res)
        if self.csvName:
            saveCsv(res, self.csvName)
            return self.psi
        else:
            return self.res, self.psi





import numpy as np
import pandas as pd

class BinomialTest:

    def __init__(self, df_res=None, ranking=None,
                 target=None, proba=None, alpha=0.05,):
        """
        主标尺下的二项检验，支持两种调用方式
        方式一（推荐）：给定样本的评级等级 ranking、预测PD proba、真实违约 target
        方式二（不推荐）：给定依据评级等级分组后 df_res (DataFrame)
        :param df_res:
        :param ranking:
        :param target:
        :param proba:
        :param alpha:
        """
        import scipy.stats as st
        self.st = st

        self.ranking = ranking
        self.target = target
        self.proba = proba
        self.alpha = alpha

        if self.target is not None:
            groups = pd.DataFrame({
                'ranking': self.ranking,
                'target': self.target,
                'proba': self.proba,
            }).groupby('ranking')

            self.res = pd.DataFrame({
                'ranking': groups.target.sum().index,
                'counts': groups.target.count(),
                'default_nums': groups.target.sum(),
                'default': groups.target.mean(),
                'proba': groups.proba.mean(),
            })
        else:
            self.res = df_res

    def binomial_pvalue(
            self, count_col='counts', proba_col='proba', default_nums_col='default_nums'
    ):
        """
        若给定等级分组后的 df_res 则需指定以下列
            样本计数列 count_col
            分级内的平均预测概率 proba_col
            违约样本数 default_num_col
        :param count_col:
        :param proba_col:
        :param default_nums_col:
        :return:
        """
        group_proba = self.res[proba_col].copy()
        group_proba[group_proba != 1] = group_proba[group_proba != 1] + 1e-6
        group_proba[group_proba == 1] = group_proba[group_proba == 1] - 1e-6

        self.res['Z'] = (self.res[default_nums_col] - self.res[count_col] * group_proba) /\
                   np.sqrt(self.res[count_col] * group_proba * (1 - group_proba))
        self.res['distr'] = self.st.norm.cdf(-np.abs(self.res.Z))
        self.res['p_value'] = self.res.distr * 2
        self.res['binomial_test'] = [True if p > self.alpha else False for p in self.res.p_value]
        return self.res

    def binomial_boundary(
            self, count_col='counts', proba_col='proba', default_nums_col='default_nums',
            unilateral=None,
    ):
        """
        若给定等级分组后的 df_res 则需指定以下列
            样本计数列 count_col
            分级内的平均预测概率 proba_col
            违约样本数 default_num_col
        :param count_col:
        :param proba_col:
        :param default_nums_col:
        :param unilateral: left right
        :return:
        """
        group_proba = self.res[proba_col].copy()
        group_proba[group_proba != 1] = group_proba[group_proba != 1] + 1e-6


        if unilateral == 'left':
            self.res['d_min'] = self.st.binom.ppf(self.alpha, self.res[count_col], group_proba)
            self.res['binomial_test'] = self.res[default_nums_col] >= self.res.d_min
        elif unilateral == 'right':
            self.res['d_max'] = self.st.binom.ppf(1 - self.alpha, self.res[count_col], group_proba)
            self.res['binomial_test'] = self.res[default_nums_col] <= self.res.d_max
        else:
            self.res['d_min'] = self.st.binom.ppf(self.alpha / 2, self.res[count_col], group_proba)
            self.res['d_max'] = self.st.binom.ppf(1 - (self.alpha / 2), self.res[count_col], group_proba)
            self.res['binomial_test'] = np.logical_and(
                self.res[default_nums_col] <= self.res.d_max, self.res[default_nums_col] >= self.res.d_min)
        return self.res

def herfindahl(ranking: pd.Series):
    """
    Herfindahl index 赫芬达尔指数
    计量样本中不同评级的集中度, 一般值域为[1/k, 1]最大为20%。
    :param ranking: 评级等级
    :return: Herfindahl index
    """
    value_counts = ranking.value_counts()
    return sum((value_counts / value_counts.sum())**2)




import pandas as pd
from scipy import stats
from itertools import combinations


def spearmanr(df, columns):
    """
    描述：spearman 相关系数。
    Spearman 秩相关系数是两个数据集之间关系单调性的非参数度量。
    与 Pearson 相关性不同，Spearman 相关性不假设两个数据集都是正态分布的。
    Spearman 系数在 -1 和 +1 之间变化，0 表示没有相关性。
    -1 或 +1 的相关性意味着精确的单调关系。正相关意味着随着 x 的增加，y 也会增加。
    负相关意味着随着 x 增加，y 减少。

    p 值粗略地表示不相关系统生成具有 Spearman 相关性的数据集的概率，
    p 值并不完全可靠，但对于大于 500 左右的数据集可能是合理的。

    :param df: DataFrame 需要统计相关系数的数据。
    :param columns: 需要统计相关系数的数据列名。
    :return: DataFrame
    """
    group = list(combinations(columns, 2))
    df_spearmanr = pd.DataFrame(group, columns=['col_a', 'col_b'])

    def one_group_spearmanr(cols):
        """"""
        return stats.spearmanr(*cols).correlation

    group_data = [[df[col_a].to_numpy(), df[col_b].to_numpy()] for col_a, col_b in group]
    df_spearmanr['corr'] = list(map(one_group_spearmanr, group_data))
    df_spearmanr = df_spearmanr.pivot(index='col_a', columns='col_b', values='corr')

    df_spearmanr = df_spearmanr.sort_index()
    df_spearmanr = df_spearmanr[sorted(df_spearmanr.columns.tolist())]
    return df_spearmanr



import numpy as np
import pandas as pd


class DefaultAnalysis:
    """
    该方法实现了对于单指标变量的违约分析，并绘制了相应的违约概率图。
    目前该方法支持对于等距等频Woe的分析。
    """

    def __init__(self, woe,
                 col=None,
                 sortBy='bad',
                 title='违约概率分析', subtitle=None,
                 imageName=None, csvName=None,
                 jupyterNoteBook=True,
                 width=None, height=None,
                 _round=4):
        """
        :param woe: 需要分析的 Woe or DataFrame
        :param sortBy: good, bad, sampleRate, badRateInBin, default='bad'
        :param title: 图表的Title
        :param subtitle: 图表的SubTitle
        :param imageName: 图片保存的名称
        :param csvName: Woe表格信息保存的名字
        :param jupyterNoteBook: 是否在返回JupyterNotebook显示对象
        :param width: 图像宽度
        :param height: 图像高度
        :param _round: 小数点位数
        """
        self.woe = woe
        self.sortBy = sortBy

        self.title = title
        self.subtitle = subtitle
        self.imageName = imageName
        self.csvName = csvName
        self.jupyterNoteBook = jupyterNoteBook
        self.width = width
        self.height = height
        self._round = _round

        self.df = woe
        self.col = col

    def __analysisWoe(self, ):
        woe = self.woe
        Samples = self.woe.total.sum()
        woe['sampleRate'] = woe.total / Samples
        woe['badRateInBin'] = woe.bad / woe.total
        woe['center'] = (woe.max_bin + woe.min_bin) / 2
        woe['scaleCenter'] = (woe.center - woe.center.min()) / \
                             (woe.center.max() - woe.center.min())
        woe['bucket'] = '[' + \
                        woe.min_bin.astype(str) + ', ' + woe.max_bin.astype(str) + ']'
        return woe

    def __analysisCate(self, ):

        samples = len(self.df)
        groupNameList = []
        goodList = []
        badList = []
        sampleRate = []
        badRateInBin = []
        countList = []

        for groupName, group in self.df.groupby(by=self.col):
            count = len(group)
            bad = group.target.sum()
            good = count - bad

            groupNameList.append(groupName)
            goodList.append(int(good))
            badList.append(int(bad))
            sampleRate.append(float(count / samples))
            badRateInBin.append(float(int(bad) / int(count)))
            countList.append(count)

        dfCate = pd.DataFrame(
            {
                'groupName': groupNameList,
                'Count': countList,
                'bad': badList,
                'good': goodList,
                'sampleRate': sampleRate,
                'badRateInBin': badRateInBin,
            }
        )
        dfCate.sort_values(by=self.sortBy, inplace=True)
        return dfCate


    def DefaultFreGraph(self, ) -> Grid:

        woe = self.__analysisWoe()

        xData = woe.Bucket.to_list()
        goodSample = (woe.total - woe.bad).to_list()
        badSample = woe.bad.to_list()
        badRateInBin = (woe.badRateInBin.round(self._round) * 100).to_list()
        sampleRate = (woe.sampleRate.round(self._round) * 100).to_list()

        totalNums = woe.total.sum()
        badNums = woe.bad.sum()
        goodNums = totalNums - badNums
        badRate = np.round(badNums / totalNums, self._round) * 100

    def DefaultCateGraph(self, ) -> Grid:

        dfCate = self.__analysisCate().round(self._round)

        xData = dfCate.groupName.to_list()
        badSample = dfCate.bad.to_list()
        goodSample = dfCate.good.to_list()
        badRateInBin = dfCate.badRateInBin.to_list()
        sampleRate = dfCate.sampleRate.to_list()

        totalNums = dfCate.Count.sum()
        goodNums = dfCate.good.sum()
        badNums = dfCate.bad.sum()
        badRate = round(badNums / totalNums, self._round)


    def DefaultCatePng(self, title='None title', sort_by=None, width=0.35, rotation=30, **kwargs):
        """

        :param title:
        :param sort_by: bad good bad_rate sample_rate
        :param width:
        :return:
        """
        import matplotlib.pyplot as plt

        dfCate = self.__analysisCate().round(self._round)

        xData = dfCate.groupName.to_list()
        badSample = dfCate.bad.to_list()
        goodSample = dfCate.good.to_list()
        badRateInBin = dfCate.badRateInBin.to_list()
        sampleRate = dfCate.sampleRate.to_list()

        totalNums = dfCate.Count.sum()
        goodNums = dfCate.good.sum()
        badNums = dfCate.bad.sum()
        badRate = round(badNums / totalNums, self._round)

        data = pd.DataFrame(
            {
                'labels': xData,
                'bad': badSample,
                'good': goodSample,
                'bad_rate': badRateInBin,
                'sample_rate': sampleRate,
            }
        )

        if sort_by:
            data.sort_values(by=sort_by, inplace=True)

        x = np.arange(len(data.labels))  # the label locations

        fig, ax = plt.subplots(2, 1, figsize=(10, 10), dpi=120)
        rects1 = ax[0].bar(x - width / 2, data.bad, width, label='bad', color='tab:red', alpha=0.7)
        rects2 = ax[0].bar(x + width / 2, data.good, width, label='good', color='tab:green', alpha=0.7)

        # Add some text for labels, title and custom x-axis tick labels, etc.
        ax[0].set_ylabel('Samples')
        if kwargs.get('font'):
            ax[0].set_title(title, fontproperties=kwargs.get('font'))
        else:
            ax[0].set_title(title)
        ax[0].set_xticks(x)
        ax[0].set_xticklabels(data.labels)
        ax[0].legend()
        ax[0].xaxis.set_tick_params(rotation=rotation, labelsize=10)

        def autolabel(rects):
            """Attach a text label above each bar in *rects*, displaying its height."""
            for rect in rects:
                height = rect.get_height()
                ax[0].annotate('{}'.format(height),
                               xy=(rect.get_x() + rect.get_width() / 2, height),
                               xytext=(0, 3),  # 3 points vertical offset
                               textcoords="offset points",
                               ha='center', va='bottom')

        line_ax = ax[1]
        _line_ax = line_ax.twinx()
        line1 = line_ax.plot(x, data.bad_rate, label='bad rate in bin', color='tab:red')
        line2 = _line_ax.plot(x, data.sample_rate, label='sample rate', color='tab:green')
        ax[1].set_xticks(x)
        ax[1].set_xticklabels(data.labels)
        ax[1].legend()
        _line_ax.legend()
        ax[1].set_ylabel('rate')
        ax[1].set_xlabel(f'Total sample: {totalNums} Good: {goodNums} Bad: {badNums} Bad rate: {badRate}')
        ax[1].xaxis.set_tick_params(rotation=rotation, labelsize=10)

        autolabel(rects1)
        autolabel(rects2)

        fig.tight_layout()


def plot_bin(
        df,
        title='None title',
        sort_by=None,
        width=0.35,
        rotation=30,
        _round=4,
        **kwargs
    ):
    """

    :param title:
    :param sort_by: bad good bad_rate sample_rate
    :param width:
    :return:
    """
    labels = df.min_bin.astype(str) + '_' + df.max_bin.astype(str)
    badSample = df.bad
    goodSample = df.total - df.bad
    badRateInBin = df.bad_rate
    sampleRate = df.total/df.total.sum()
    totalNums = df.total.sum()
    goodNums = goodSample.sum()
    badNums = df.bad.sum()
    badRate = round(badNums / totalNums, _round)

    import matplotlib.pyplot as plt

    data = pd.DataFrame(
        {
            'labels': labels,
            'bad': badSample,
            'good': goodSample,
            'bad_rate': badRateInBin,
            'sample_rate': sampleRate,
        }
    )

    if sort_by:
        data.sort_values(by=sort_by, inplace=True)

    x = np.arange(len(data.labels))  # the label locations

    fig, ax = plt.subplots(2, 1, figsize=(10, 10), dpi=120)
    rects1 = ax[0].bar(x - width / 2, data.bad, width, label='bad', color='tab:red', alpha=0.7)
    rects2 = ax[0].bar(x + width / 2, data.good, width, label='good', color='tab:green', alpha=0.7)

    # Add some text for labels, title and custom x-axis tick labels, etc.
    ax[0].set_ylabel('Samples')
    if kwargs.get('font'):
        ax[0].set_title(title, fontproperties=kwargs.get('font'))
    else:
        ax[0].set_title(title)
    ax[0].set_xticks(x)
    ax[0].set_xticklabels(data.labels)
    ax[0].legend()
    ax[0].xaxis.set_tick_params(rotation=rotation, labelsize=10)

    def autolabel(rects):
        """Attach a text label above each bar in *rects*, displaying its height."""
        for rect in rects:
            height = rect.get_height()
            ax[0].annotate('{}'.format(height),
                           xy=(rect.get_x() + rect.get_width() / 2, height),
                           xytext=(0, 3),  # 3 points vertical offset
                           textcoords="offset points",
                           ha='center', va='bottom')

    line_ax = ax[1]
    _line_ax = line_ax.twinx()
    line1 = line_ax.plot(x, data.bad_rate, label='bad rate in bin', color='tab:red')
    line2 = _line_ax.plot(x, data.sample_rate, label='sample rate', color='tab:green')
    ax[1].set_xticks(x)
    ax[1].set_xticklabels(data.labels)
    ax[1].legend()
    _line_ax.legend()
    ax[1].set_ylabel('rate')
    ax[1].set_xlabel(f'Total sample: {totalNums} Good: {goodNums} Bad: {badNums} Bad rate: {badRate}')
    ax[1].xaxis.set_tick_params(rotation=rotation, labelsize=10)
    autolabel(rects1)
    autolabel(rects2)
    fig.tight_layout()
    return None



from .woe_frequency import *
from .woe_distance import *
from .woe_category import *
from .woe_manually import *
from .woe_pearson import *
from .woe_chi2 import *
from .woe_tree import *
from .woe_ks import *
from .woe_kmeans import *
from typing import Union, Type


TypeWoe = Union[
    Type[DistanceBins],
    Type[FrequencyBins],
    Type[CategoryBins],
    Type[ManuallyBins],
    Type[PearsonBins],
    Type[Chi2Bins],
    Type[TreeBins],
    Type[KSBins],
    Type[KmeansBins],
]



import numpy as np
import pandas as pd
from ..utils.operator import NumRange

def PDToScore(PD, base_score=300, pdo=17, log='log2'):
    """
    描述：将预测概率转换为分数，这里区分以e为底取log或以2为底取log。
    以2为底取log，B 即为 pdo
    以e为底取log，B 即为 pdo / ln(2)
    1. ln(odds) = ln(p / 1 - p) = WX^T + b
    2. score = A - B ln(odds)
             = A - B ln(p / 1 - p)
             = A - B (WX^T + b)
    2. score = A - B log2(odds)
             = A - B log2(p / 1 - p)
             = A - B (WX^T + b)
    参数：
    :param PD: 违约概率
    :param base_score: 基准分
    :param pod: Points to Duble the Odds, Odds（好坏比）变为2倍时，所减少的信用分。
    :return: 分数
    """
    if log == 'log2':
        log = np.log2
    else:
        log = np.log
    return base_score + pdo * log((1 - PD) / PD)

def BaseScoreAndPDO(odds=1/50, base_score=600, pdo=20):
    """
    计算 base_score, pdo

    参数：
    :param odds:
    :param base_score:
    :param pdo:
    :return: A, B

    示例：
    >>> BaseScoreAndPDO()
    """
    B = pdo / np.log(2)
    A = base_score + B * np.log(odds)
    return A, B

def model_score_by_weight(
        df_woeval, df_param, weight_name='weight',
        target_name='target', reset_index=True, score_border=None):
    """
    描述：使用 logistics 回归的系数作为权重，返回模型分数。

    :param df_woeval: 指标经过 WOE 转换的分数值。
    :param df_param: 指标的权重。
    :param weight_name: 权重的名称。
    :param target_name: 标签的名称。
    :param reset_index: 重置 index。
    :param score_border: 分值的值域。
    :return:
    """

    df_score = pd.DataFrame()
    columns = df_param.index.tolist()

    if isinstance(df_param, pd.DataFrame):
        weight = np.array(df_param[weight_name].tolist())
    elif isinstance(df_param, pd.Series):
        weight = np.array(df_param.tolist())
    else:
        weight = [0] * len(df_param)

    weight = weight / weight.sum()
    score = df_woeval[columns].dot(weight)

    if score_border:
        a, b = score_border
        X_min, X_max = score.min(), score.max()
        df_score['score'] = NumRange(a, b, X_min, X_max, score)
    else:
        df_score['score'] = score

    df_score[target_name] = df_woeval[target_name]

    if reset_index:
        df_score.reset_index(inplace=True, drop=True)

    return df_score




import numpy as np
import pandas as pd
from pandas import DataFrame, Series
import statsmodels.api as sm
from typing import Any, List, Dict, Tuple, Callable, Literal, Optional

from ..utils.utils import dispatch_tqdm
from ..utils.operator import Div
from ..utils.logger import LoggerMixin
from ..evaluate.binary_metrics import ks_test, BinaryMetrics, Chi2
from .schema import *


__all__ = [
    "FeatureEvaluate",
]


class FeatureEvaluate(LoggerMixin):
    """"""
    df_report: DataFrame
    report_data: List[Dict[str, Any]]

    def __init__(
            self,
            df: DataFrame,
            target_name: str = 'target',
            columns: Optional[List[str]] = None,
            columns_mapping: Optional[Dict[Any, str]] = None,
            eval_method: Optional[Literal["origin", "woe"]] = "origin",
            logger: Optional[Callable] = None,
            verbose: Optional[bool] = True,
            *args: Any,
            **kwargs: Any,
    ):
        """
        :param df: DataFrame
        :param target_name: str
        :param columns: List[str]
        :param columns_mapping: Dict[Any, str]
        :param eval_method: Literal["origin", "woe"]
        :param logger: Callable
        :param verbose: bool
        """
        self.df = df
        self.target_name = target_name
        self.columns_mapping = columns_mapping
        self.eval_method = eval_method
        self.logger = logger
        self.verbose = verbose
        self.args = args
        self.kwargs = kwargs
        self._validate_data()
        self._analysis_columns(columns)
        self._base_stats()
        self.output_fields = [
            'var_name', 'desc', 'bad_rate', 'missing', 'miss_bad_rate', 'not_miss_bad_rate',
            'null_lift', 'not_null_lift', 'min', 'max', 'std', 'mode', 'skew', 'kurt',
            'cv', 'unique', 'bad', 'count', 'good', 'ks_test', 'R2', 'intercept',
            'pvalue_intercept', 'param', 'pvalue_param', 'auc', 'ks',
            'chi_statistic', 'chi_pvalue'
        ]

    def __call__(self, *args, **kwargs):
        """"""

    def _div(self, a, b, else_value=-9999.):
        """"""
        if b != 0:
            return a / b
        else:
            return else_value

    def _validate_data(self):
        """"""
        if not isinstance(self.df, DataFrame):
            raise TypeError(f'df must be a pandas DataFrame, got {type(self.df)}')
        if self.target_name not in self.df.columns:
            raise ValueError(f'target_name {self.target_name} not in df.columns')

    def _analysis_columns(self, columns: Optional[List[str]] = None):
        """"""
        if columns is None:
            columns = self.df.columns.to_list()
            for drop_col in [self.target_name]:
                if drop_col in columns:
                    columns.remove(drop_col)
            self.columns = columns
        else:
            self.columns = columns

    def _prepare_data(self, column: Series, target: Series) -> Tuple[Series, Series]:
        """"""
        condition = ~(column.isna() | (np.isinf(column.astype('float'))))
        column_data = column[condition].copy()
        target_data = target[condition].copy()
        return column_data, target_data

    def _base_stats(self):
        """"""
        self.bad_rate = self.df[self.target_name].mean()

    def _column_info(self, column: Series, report: FeatureEvaluateSchema):
        """"""
        report.var_name = column.name
        report.bad_rate = self.bad_rate
        if self.columns_mapping:
            report.desc = self.columns_mapping.get(report.var_name)

    def _column_base(self, column_data: Series, target_data: Series, report: FeatureEvaluateSchema):
        """"""
        report.min = float(column_data.min())
        report.max = float(column_data.max())
        report.std = float(column_data.std())
        report.mode = float(column_data.mode(dropna=True).tolist()[0])
        report.skew = float(column_data.skew())
        report.kurt = float(column_data.kurt())
        report.cv = Div.div(float(column_data.std()), float(column_data.mean()))
        report.unique = len(column_data.unique())
        report.bad = int(target_data.sum())
        report.count = int(target_data.count())
        report.good = int(report.count - report.bad)
        report.ks_test = ks_test(column_data, target_data)[0]

        # chi2
        chi_statistic, chi_pvalue = Chi2(
            column_data.values.reshape(-1, 1),
            target_data.values)
        report.chi_statistic = chi_statistic
        report.chi_pvalue = chi_pvalue

    def _column_missing(self, column: Series, target: Series, report: FeatureEvaluateSchema):
        """"""
        report.missing = column.isna().mean()
        report.miss_bad_rate = target[column.isna()].mean()
        report.not_miss_bad_rate = target[column.notna()].mean()
        report.null_lift = self._div(report.miss_bad_rate, self.bad_rate)
        report.not_null_lift = self._div(report.not_miss_bad_rate, self.bad_rate)

    def _column_logist(
            self,
            column: Optional[Series] = None,
            target: Optional[Series] = None,
            column_data: Optional[Series] = None,
            target_data: Optional[Series] = None,
            report: Optional[FeatureEvaluateSchema] = None,
    ) -> FeatureEvaluateSchema:
        """"""
        if self.eval_method == "origin":
            x_data = column_data
            y_data = target_data
        elif self.eval_method == "woe":
            x_data = column
            y_data = target
        else:
            raise ValueError(f"eval_method {self.eval_method} not support")
        x_data = sm.add_constant(x_data, has_constant='add')
        try:
            model = sm.Logit(y_data, x_data).fit(disp=0)
            report.R2 = model.prsquared
            report.intercept = model.params.const
            report.pvalue_intercept = model.pvalues.const
            report.param = model.params[report.var_name]
            report.pvalue_param = model.pvalues[report.var_name]

            binary_metrics = BinaryMetrics(target=model.model.endog, proba=model.predict())
            metrics = binary_metrics.evaluate()
            report = report.model_copy(update=metrics.to_univariate())
        except Exception as e:
            self._logger(msg=f'Model Error! Feature Name: {report.var_name}, Error Msg: {e}', color="red")
        return report

    def stats(self) -> DataFrame:
        """"""
        eval_report = []
        for column_name in dispatch_tqdm(self.columns, desc="FeatureEvaluate"):
            feature_report = FeatureEvaluateSchema()
            column = self.df[column_name]
            column_data, target_data = self._prepare_data(column=column, target=self.df[self.target_name])
            self._column_info(column=column, report=feature_report)
            self._column_missing(column=column, target=self.df[self.target_name], report=feature_report)
            self._column_base(column_data=column_data, target_data=target_data, report=feature_report)
            feature_report = self._column_logist(
                column=column, target=self.df[self.target_name],
                column_data=column_data, target_data=target_data,
                report=feature_report
            )
            eval_report.append(feature_report.model_dump())
        return pd.DataFrame(eval_report)[self.output_fields]



try:
    from sklearn.tree import DecisionTreeClassifier
except ImportError:
    raise ImportError("Please install scikit-learn to use this module. pip install scikit-learn")
import pandas as pd
from pandas import Series, DataFrame
from typing import Any, List, Union, Optional
from kivi.woe.base import WOEMixin


__all__ = ['TreeBins']


class TreeBins(WOEMixin):
    """决策树分箱"""
    cutoff_point: List[float]

    def __init__(
            self,
            variables: Series,
            target: Series,
            bins: Optional[int] = 5,
            fill_bin: Optional[bool] = True,
            var_name: Optional[str] = None,
            abnormal_vals: Optional[List[Union[str, int, float]]] = None,
            weight: Optional[Any] = None,
            dtc_weight: Optional[Any] = None,
            decimal: Optional[int] = 6,
            *args: Any,
            **kwargs: Any,
    ):
        """
        描述：决策树分箱方法，使用决策树最优 entropy 进行分箱分析。

        :param variables: 待分箱变量
        :param target: 目标标签变量
        :param bins: 决策树分箱中最大的叶子结点数量，一般对应的是最终分箱数量，默认为 5 。
        :param abnormal_values: 特殊值分箱，在变量存在特殊值时单独分一箱，如 -1111, -9999。
        :param fill_bin: 在各分箱中偶发性会出现 good 或 bad 为 0 的情况，默认 fill_pos 为 True ，为该分箱填充 0.5。

        Example:
            woe = TreeBins(variables, target, bins=5, fill_bin=True)
            woe.fit()
        """
        self.variables = variables
        self.target = target
        self.bins = bins
        self.max_leaf_nodes = bins
        self.fill_bin = fill_bin
        self.abnormal_vals = abnormal_vals
        self.dtc_weight = dtc_weight
        self.decimal = decimal
        self.args = args
        self.kwargs = kwargs
        self.data_prepare(
            variables=variables, target=target,
            weight=weight, var_name=var_name)

    def tree_cutoff_point(self, ):
        """ """
        clf = DecisionTreeClassifier(
            criterion='entropy', max_leaf_nodes=self.max_leaf_nodes, min_samples_leaf=0.05)
        clf.fit(self.variables.to_numpy().reshape(-1, 1), self.target.to_numpy().reshape(-1, 1), sample_weight=self.dtc_weight)
        n_nodes = clf.tree_.node_count
        left, right = clf.tree_.children_left, clf.tree_.children_right
        threshold = clf.tree_.threshold
        self.cutoff_point = sorted([-float('inf')] + [threshold[i] for i in range(n_nodes) if left[i] != right[i]] + [float('inf')])

    def fit(
            self,
            score: Optional[bool] = True,
            origin_border: Optional[bool] = False,
            order: Optional[bool] = True,
            **kwargs: Any,
    ) -> DataFrame:
        """

        :param score: 是否增加 WOEMixin score。
        :param origin_border: 是否增加 分箱中的最大值与最小值。
        :param order: 是否增加单调性判断。
        :return: DataFrame WOEMixin result.
        """
        self.tree_cutoff_point()
        _bucket, _bins = pd.cut(self.df_data.variables, self.cutoff_point, include_lowest=True, retbins=True)
        if self.weight is None:
            bucket = pd.DataFrame({
                'variables': self.df_data.variables,
                'target': self.df_data.target,
                'bucket': _bucket,
            }).groupby('bucket', as_index=True, observed=False)
        else:
            bucket = self.get_weighted_buckets(
                value_cut=_bucket, variables=self.df_data.variables, target=self.df_data.target, weight=self.df_data.weight)
        self.cal_woe_iv(bucket, score=score, origin_border=origin_border, order=order)
        return self.woe



import numpy as np
from sklearn.linear_model import LassoCV
import matplotlib.pyplot as plt
from pandas import DataFrame, Series
from sklearn.metrics import r2_score, mean_squared_error
from sklearn.compose import TransformedTargetRegressor



class Lasso:

    def __init__(self, data, y_name, alphas, normalize=False, random_state=None, log_y=False):
        """
        Des: Lasso 回归方法
        :param data: 数据集
        :param y_name: y 名称
        :param alphas: 待选alpha参数
        :param normalize: 是否正则化，默认False
        :param random_state: 随机种子
        :param log_y: 是否y取对数，默认False
        """
        y = data.pop(y_name)
        self.exog = data.copy()
        self.endog = y.copy()

        self.model = self.lasso(self.exog, self.endog, alphas, normalize, random_state, log_y)
        if log_y:
            self.coef = Series(self.model.regressor_.coef_, index=self.exog.columns).abs().sort_values(ascending=False)
            self.intercept = self.model.regressor_.intercept_
            self.alpha = self.model.regressor_.alpha_

        else:
            self.coef = Series(self.model.coef_, index=self.exog.columns).abs().sort_values(ascending=False)
            self.intercept = self.model.intercept_

        y_pre = self.model.predict(self.exog)
        self.mse = mean_squared_error(self.endog, y_pre)
        self.r2_score = r2_score(self.endog, y_pre)

    def lasso(self, X, y, alphas, normalize, random_state, log_y):
        """
        Des: Lasso 回归
        :param X:
        :param y:
        :param alphas:
        :param normalize:
        :param random_state:
        :param log_y:
        :return:
        """
        if log_y:
            _model = LassoCV(alphas=alphas, normalize=normalize)
            tt_model = TransformedTargetRegressor(
                regressor=_model,
                func=np.log,
                inverse_func=np.exp
            )
            return tt_model.fit(X, y)
        else:
            return LassoCV(alphas=alphas, normalize=normalize, random_state=random_state).fit(X, y)

    def plot_variable_importance(self, ):
        """
        Des: 绘制参数重要性图
        :return:
        """
        plt.figure(dpi=80)
        plt.bar(self.coef.index, self.coef.values)
        return None

    def select_features(self, feature_nums=None, min_treshold=0.1):
        """
        Des: 选择最重要的几个特征
        :param feature_nums: 选择特征数
        :param min_treshold: 特征重要性阈值
        :return: 相应特征的数据
        """
        if feature_nums >= self.exog.shape[-1]:
            raise Exception(ValueError, 'feature_nums must be less than exog columns!')

        try:
            from sklearn.feature_selection import SelectFromModel
            sfm = SelectFromModel(self.model, threshold=min_treshold, prefit=True)
            n_features = sfm.transform(self.exog).shape[1]

            while n_features > feature_nums:
                sfm.threshold += 0.1
                X_transform = sfm.transform(self.exog)
                n_features = X_transform.shape[1]
            if feature_nums != n_features:
                return self.exog[self.coef.index[: feature_nums]]
            return DataFrame(X_transform, columns=self.coef.index[: feature_nums])

        except:
            return self.exog[self.coef.index[: feature_nums]]



class __ConnectSQL__:

    def __init__(self, dataBase, ConnectInfo):
        self.dataBase = dataBase
        self.ConnectInfo = ConnectInfo
        self.ConnectInfo['dataBase'] = dataBase

    def __connectInfo__(self, dataBaseType, name, password, host, prot, dataBase):
        if dataBaseType == 'MySql':
            CONNECT = 'mysql+pymysql://{}:{}@{}:{}/{}?charset=UTF8MB4'.format(
                name, password, host, prot, dataBase)
            return CONNECT
        if dataBaseType == 'Oracle':
            CONNECT = 'oracle+cx_oracle://{}:{}@{}:{}/{}'.format(
                name, password, host, prot, dataBase)
            return CONNECT

    def engine(self, ):
        from sqlalchemy import create_engine
        CONNECT = self.__connectInfo__(**self.ConnectInfo)
        engin = create_engine(CONNECT)
        print('Connect to {} success!'.format(CONNECT))
        return engin


class ShowDataBase:

    def __init__(self, engin):
        from sqlalchemy import inspect
        self.inspector = inspect(engin)

    def showTable(self, ):
        tablesName = self.inspector.get_table_names()
        return tablesName

    def showColumns(self, tableName):
        columns = [
            col['name'] for col in self.inspector.get_columns(tableName)
        ]
        return columns


def GetSqlEngine(dataBaseName, connectInfo):
    """
    :param dataBaseName: dataBaseName
    :param connectInfo: e.g.
        connectInfo = {
            'dataBaseType': 'MySql' or 'Oracle',
            'name': 'root',
            'password': '********',
            'host': 'localhost',
            'prot': '3306',
        }
    :return: engine
    """
    connectSql = __ConnectSQL__(dataBaseName, connectInfo)
    return connectSql.engine()


class MySql:
    def __init__(self, host='localhost', port='3306', user='root', password='1702', spark=None):
        self.host = host
        self.port = port
        self.spark = spark
        self.prop = {
            'user': user,
            'password': password,
        }
        
    def getPyUrl(self, db):
        """
        获取Pandas URL
        """
        url = 'mysql+pymysql://{}:{}@{}:{}/{}?charset=UTF8MB4'.format(
            self.prop.get('user'), self.prop.get('password'), 
            self.host, self.port, db)
        return url
    
    def getEngin(self, db):
        """
        获取Pandas URL链接引擎
        """
        from sqlalchemy import create_engine
        return create_engine(self.getPyUrl(db))
        
    def getSparkUrl(self, db):
        """
        获取PySpark DataFrame URL链接引擎
        """
        url = f'jdbc:mysql://{self.host}:{self.port}/{db}'
        return url
    
    def readMySql(self, db, table):
        """
        读取MySQL至PySpark DataFrame
        """
        url = self.getSparkUrl(db)
        df = self.spark.read.jdbc(url=url, table=table, properties=self.prop)
        return df
    
    def writeTable(self, df, db, table_name, mode='error'):
        """
        描述：保存PySpark DataFrame到MySQL
        
        参数：
        :param df:
        :param db:
        :param table_name:
        :param mode:
            * ``append``: Append contents of this :class:`DataFrame` to existing data.
            * ``overwrite``: Overwrite existing data.
            * ``ignore``: Silently ignore this operation if data already exists.
            * ``error`` or ``errorifexists`` (default case): Throw an exception if data already   
        """
        url = self.getSparkUrl(db)
        df.write.jdbc(url=url, table=table_name, mode=mode, properties=self.prop)
        return None
    

    
def dropTable(table_name, con):
    """
    删除表
    """
    from sqlalchemy.orm import sessionmaker
    session = sessionmaker(con)
    tables = ShowDataBase(con).showTable()
    if table_name in tables:
        print("查询到表：{}。".format(table_name))
        confirm = input('确认删除请输入‘y’: ')
    else:
        print("查询到表：{} 不存在。".format(table_name))
    if confirm == 'y':
        with con.connect() as connection:
            with session(bind=connection) as sess:
                sess.execute("DROP TABLE IF EXISTS {}".format(table_name))
        print('Droped Table: {}'.format(table_name))
    return None


def checkTableExists():
    """
    描述：检查表是否存在
    """
    return None

def renameTable():
    """
    描述：对表进行重命名
    """
    return None

